# pylint: disable=unused-import,attribute-defined-outside-init
from typing import Callable, Iterable, Dict, List, Union, Tuple, Type
from types import FunctionType
from abc import ABC
import inspect
import numpy as np

from nmmo.task.group import Group
from nmmo.task.game_state import GameState
from nmmo.task.predicate_api import Predicate, make_predicate, arg_to_string
from nmmo.task import base_predicates as bp

class Task(ABC):
  """ A task is used to calculate rewards for agents in assignee
      based on the predicate and game state
  """
  def __init__(self,
               eval_fn: Callable,
               assignee: Union[Iterable[int], int],
               reward_multiplier = 1.0,
               embedding = None,
               spec_name: str = None):
    if isinstance(assignee, int):
      self._assignee = (assignee,)
    else:
      assert len(assignee) > 0, "Assignee cannot be empty"
      self._assignee = tuple(set(assignee)) # dedup
    self._eval_fn = eval_fn
    self._reward_multiplier = reward_multiplier
    self._embedding = None if embedding is None else np.array(embedding, dtype=np.float16)
    self.spec_name = spec_name # None if not created using TaskSpec
    self.name = self._make_name(self.__class__.__name__,
                                eval_fn=eval_fn, assignee=self._assignee)
    self.reset()

  def reset(self):
    self._stop_eval = False
    self._last_eval_tick = None
    self._progress = 0.0
    self._completed_tick = None
    self._max_progress = 0.0
    self._positive_reward_count = 0
    self._negative_reward_count = 0

  def close(self):
    if self._stop_eval is False:
      if isinstance(self._eval_fn, Predicate):
        self._eval_fn.close()
      self._stop_eval = True

  @property
  def assignee(self) -> Tuple[int]:
    return self._assignee

  @property
  def completed(self) -> bool:
    return self._completed_tick is not None

  @property
  def reward_multiplier(self) -> float:
    return self._reward_multiplier

  @property
  def reward_signal_count(self) -> int:
    return self._positive_reward_count + self._negative_reward_count

  @property
  def embedding(self):
    return self._embedding

  def set_embedding(self, embedding):
    self._embedding = embedding

  def _map_progress_to_reward(self, gs: GameState) -> float:
    """ The default reward is the diff between the old and new progress.
        Once the task is completed, no more reward is provided.

        Override this function to create a custom reward function
    """
    if self.completed:
      return 0.0

    new_progress = max(min(self._eval_fn(gs)*1.0,1.0),0.0)
    diff = new_progress - self._progress
    self._progress = new_progress
    if self._progress >= 1:
      self._completed_tick = gs.current_tick
      diff = 1.0  # give out the max reward when task is completed

    return diff

  def compute_rewards(self, gs: GameState) -> Tuple[Dict[int, float], Dict[int, Dict]]:
    """ Environment facing API

    Returns rewards and infos for all agents in subject
    """
    reward = self._map_progress_to_reward(gs) * self._reward_multiplier
    self._last_eval_tick = gs.current_tick
    self._max_progress = max(self._max_progress, self._progress)
    self._positive_reward_count += int(reward > 0)
    self._negative_reward_count += int(reward < 0)
    rewards = {int(ent_id): reward for ent_id in self._assignee}
    infos = {int(ent_id): {"task_spec": self.spec_name,
                           "reward": reward,
                           "progress": self._progress,
                           "completed": self.completed}
             for ent_id in self._assignee}

    # NOTE: tasks do not know whether assignee agents are alive or dead
    #   so the Env must check it before filling in rewards and infos
    return rewards, infos

  def _make_name(self, class_name, **kwargs) -> str:
    name = [class_name] + \
      [f"{arg_to_string(key)}:{arg_to_string(arg)}" for key, arg in kwargs.items()]
    name = "("+"_".join(name).replace(" ", "")+")"
    return name

  def __str__(self):
    return self.name

  @property
  def subject(self):
    if isinstance(self._eval_fn, Predicate):
      return self._eval_fn.subject.agents
    return self.assignee

  def get_source_code(self):
    if isinstance(self._eval_fn, Predicate):
      return self._eval_fn.get_source_code()
    return inspect.getsource(self._eval_fn).strip()

  def get_signature(self):
    if isinstance(self._eval_fn, Predicate):
      return self._eval_fn.get_signature()
    signature = inspect.signature(self._eval_fn)
    return list(signature.parameters)

  @property
  def args(self):
    if isinstance(self._eval_fn, Predicate):
      return self._eval_fn.args
    # the function _eval_fn must only take gs
    return []

  @property
  def kwargs(self):
    if isinstance(self._eval_fn, Predicate):
      return self._eval_fn.kwargs
    # the function _eval_fn must only take gs
    return {}

  @property
  def progress_info(self):
    return {
      "task_spec_name": self.spec_name,
      "last_eval_tick": self._last_eval_tick,
      "completed": self.completed,
      "completed_tick": self._completed_tick,
      "max_progress": self._max_progress,
      "positive_reward_count": self._positive_reward_count,
      "negative_reward_count": self._negative_reward_count,
      "reward_signal_count": self.reward_signal_count,
    }

class OngoingTask(Task):
  def _map_progress_to_reward(self, gs: GameState) -> float:
    """Keep returning the progress reward after the task is completed.
       However, this task tracks the completion status in the same manner.
    """
    self._progress = max(min(self._eval_fn(gs)*1.0,1.0),0.0)
    if self._progress >= 1 and self._completed_tick is None:
      self._completed_tick = gs.current_tick
    return self._progress

class HoldDurationTask(Task):
  def __init__(self,
               eval_fn: Callable,
               assignee: Union[Iterable[int], int],
               hold_duration: int,
               **kwargs):
    super().__init__(eval_fn, assignee, **kwargs)
    self._hold_duration = hold_duration
    self._reset_timer()

  def _reset_timer(self):
    self._timer = 0
    self._last_success_tick = 0

  def reset(self):
    super().reset()
    self._reset_timer()

  def _map_progress_to_reward(self, gs: GameState) -> float:
    # pylint: disable=attribute-defined-outside-init
    if self.completed:
      return 0.0

    curr_eval = max(min(self._eval_fn(gs)*1.0,1.0),0.0)
    if curr_eval < 1:
      self._reset_timer()
    else:
      self._timer += 1
      self._last_success_tick = gs.current_tick

    new_progress = self._timer / self._hold_duration
    diff = new_progress - self._progress
    self._progress = new_progress
    if self._progress >= 1 and self._completed_tick is None:
      self._completed_tick = gs.current_tick
      diff = 1.0  # give out the max reward when task is completed

    return diff

######################################################################

# The same task is assigned each agent in agent_list individually
#   with the agent as the predicate subject and task assignee
def make_same_task(pred_cls: Union[Type[Predicate], Callable],
                   agent_list: Iterable[int],
                   pred_kwargs=None,
                   task_cls: Type[Task]=Task,
                   task_kwargs=None) -> List[Task]:
  # if a function is provided, make it a predicate class
  if isinstance(pred_cls, FunctionType):
    pred_cls = make_predicate(pred_cls)
  if pred_kwargs is None:
    pred_kwargs = {}
  if task_kwargs is None:
    task_kwargs = {}

  task_list = []
  for agent_id in agent_list:
    predicate = pred_cls(Group(agent_id), **pred_kwargs)
    task_list.append(predicate.create_task(task_cls=task_cls, **task_kwargs))
  return task_list

def nmmo_default_task(agent_list: Iterable[int], test_mode=None) -> List[Task]:
  # (almost) no overhead in env._compute_rewards()
  if test_mode == "no_task":
    return []

  # eval function on Predicate class, but does not use Group during eval
  if test_mode == "dummy_eval_fn":
    # pylint: disable=unused-argument
    return make_same_task(lambda gs, subject: True, agent_list, task_cls=OngoingTask)

  # the default is to use the predicate class
  return make_same_task(bp.StayAlive, agent_list, task_cls=OngoingTask)



from __future__ import annotations
from collections import deque
import math

from typing import Dict

from nmmo.systems.item import Item, Stack
from nmmo.lib.log import EventCode

"""
The Exchange class is a simulation of an in-game item exchange.
It has several methods that allow players to list items for sale,
buy items, and remove expired listings.

The _list_item() method is used to add a new item to the
exchange, and the unlist_item() method is used to remove
an item from the exchange. The step() method is used to
regularly check and remove expired listings.

The sell() method allows a player to sell an item, and the buy() method
allows a player to purchase an item. The packet property returns a
dictionary that contains information about the items currently being
sold on the exchange, such as the maximum and minimum price,
the average price, and the total supply of the items.

"""
class ItemListing:
  def __init__(self, item: Item, seller, price: int, tick: int):
    self.item = item
    self.seller = seller
    self.price = price
    self.tick = tick

class Exchange:
  def __init__(self, realm):
    self._listings_queue: deque[(int, int)] = deque() # (item_id, tick)
    self._item_listings: Dict[int, ItemListing] = {}
    self._realm = realm
    self._config = realm.config

  def _list_item(self, item: Item, seller, price: int, tick: int):
    item.listed_price.update(price)
    self._item_listings[item.id.val] = ItemListing(item, seller, price, tick)
    self._listings_queue.append((item.id.val, tick))

  def unlist_item(self, item: Item):
    if item.id.val in self._item_listings:
      self._unlist_item(item.id.val)

  def _unlist_item(self, item_id: int):
    item = self._item_listings.pop(item_id).item
    item.listed_price.update(0)

  def step(self, current_tick: int):
    """
    Remove expired listings from the exchange's listings queue
    and item listings dictionary. It takes in one parameter,
    current_tick, which is the current time in the game.

    The method starts by checking the oldest listing in the listings
    queue using a while loop. If the current tick minus the
    listing tick is less than or equal to the EXCHANGE_LISTING_DURATION
    in the realm's configuration, the method breaks out of
    the loop as the oldest listing has not expired.
    If the oldest listing has expired, the method removes it from the
    listings queue and the item listings dictionary.

    It then checks if the actual listing still exists and that
    it is indeed expired. If it does exist and is expired,
    it calls the _unlist_item method to remove the listing and update
    the item's listed price. The process repeats until all expired listings
    are removed from the queue and dictionary.
    """

    # Remove expired listings
    while self._listings_queue:
      (item_id, listing_tick) = self._listings_queue[0]
      if current_tick - listing_tick <= self._config.EXCHANGE_LISTING_DURATION:
        # Oldest listing has not expired
        break

      # Remove expired listing from queue
      self._listings_queue.popleft()

      # The actual listing might have been refreshed and is newer than the queue record.
      # Or it might have already been removed.
      listing = self._item_listings.get(item_id)
      if listing is not None and \
        current_tick - listing.tick > self._config.EXCHANGE_LISTING_DURATION:
        self._unlist_item(item_id)

  def sell(self, seller, item: Item, price: int, tick: int):
    assert isinstance(
        item, object), f'{item} for sale is not an Item instance'
    assert item in seller.inventory, f'{item} for sale is not in {seller} inventory'
    assert item.quantity.val > 0, f'{item} for sale has quantity {item.quantity.val}'
    assert item.listed_price.val == 0, 'Item is already listed'
    assert item.equipped.val == 0, 'Item has been equiped so cannot be listed'
    assert price > 0, 'Price must be larger than 0'

    self._list_item(item, seller, price, tick)

    self._realm.event_log.record(EventCode.LIST_ITEM, seller, item=item, price=price)

    self._realm.log_milestone(f'Sell_{item.__class__.__name__}', item.level.val,
      f'EXCHANGE: Offered level {item.level.val} {item.__class__.__name__} for {price} gold',
      tags={"player_id": seller.ent_id})

  def buy(self, buyer, item: Item):
    assert item.quantity.val > 0, f'{item} purchase has quantity {item.quantity.val}'
    assert item.equipped.val == 0, 'Listed item must not be equipped'
    assert buyer.gold.val >= item.listed_price.val, 'Buyer does not have enough gold'
    assert buyer.ent_id != item.owner_id.val, 'One cannot buy their own items'

    if not buyer.inventory.space:
      if isinstance(item, Stack):
        if not buyer.inventory.has_stack(item.signature):
          # no ammo stack with the same signature, so cannot buy
          return
      else: # no space, and item is not ammo stack, so cannot buy
        return

    # item is not in the listing (perhaps bought by other)
    if item.id.val not in self._item_listings:
      return

    listing = self._item_listings[item.id.val]
    price = item.listed_price.val

    self.unlist_item(item)
    listing.seller.inventory.remove(item)
    buyer.inventory.receive(item)
    buyer.gold.decrement(price)
    listing.seller.gold.increment(price)

    # TODO(kywch): tidy up the logs - milestone, event, etc ...
    #self._realm.log_milestone(f'Buy_{item.__name__}', item.level.val)
    #self._realm.log_milestone('Transaction_Amount', item.listed_price.val)
    self._realm.event_log.record(EventCode.BUY_ITEM, buyer, item=item, price=price)
    self._realm.event_log.record(EventCode.EARN_GOLD, listing.seller, amount=price)

  @property
  def packet(self):
    packet = {}
    for listing in self._item_listings.values():
      item = listing.item
      key = f'{item.__class__.__name__}_{item.level.val}'
      max_price = max(packet.get(key, {}).get('max_price', -math.inf), listing.price)
      min_price = min(packet.get(key, {}).get('min_price', math.inf), listing.price)
      supply = packet.get(key, {}).get('supply', 0) + item.quantity.val

      packet[key] = {
        'max_price': max_price,
        'min_price': min_price,
        'price': (max_price + min_price) / 2,
        'supply': supply
      }

    return packet



from functools import lru_cache, cached_property

import numpy as np

from nmmo.core.tile import TileState
from nmmo.entity.entity import EntityState
from nmmo.systems.item import ItemState
import nmmo.systems.item as item_system
from nmmo.core import action
from nmmo.lib import material, utils


class BasicObs:
  def __init__(self, values, id_col):
    self.values = values
    self.ids = values[:, id_col]

  @cached_property
  def len(self):
    return len(self.ids)

  def id(self, i):
    return self.ids[i] if i < self.len else None

  def index(self, val):
    return np.nonzero(self.ids == val)[0][0] if val in self.ids else None


class InventoryObs(BasicObs):
  def __init__(self, values, id_col):
    super().__init__(values, id_col)
    self.inv_type = self.values[:,ItemState.State.attr_name_to_col["type_id"]]
    self.inv_level = self.values[:,ItemState.State.attr_name_to_col["level"]]

  def sig(self, item: item_system.Item, level: int):
    idx = np.nonzero((self.inv_type == item.ITEM_TYPE_ID) & (self.inv_level == level))[0]
    return idx[0] if len(idx) else None


class Observation:
  def __init__(self,
    config,
    current_tick: int,
    agent_id: int,
    task_embedding,
    tiles,
    entities,
    inventory,
    market) -> None:

    self.config = config
    self.current_tick = current_tick
    self.agent_id = agent_id
    self.task_embedding = task_embedding

    self.tiles = tiles[0:config.MAP_N_OBS]
    self.entities = BasicObs(entities[0:config.PLAYER_N_OBS],
                              EntityState.State.attr_name_to_col["id"])

    self.dummy_obs = self.agent() is None
    if config.COMBAT_SYSTEM_ENABLED and not self.dummy_obs:
      latest_combat_tick = self.agent().latest_combat_tick
      self.agent_in_combat = False if latest_combat_tick == 0 else \
        (current_tick - latest_combat_tick) < config.COMBAT_STATUS_DURATION
    else:
      self.agent_in_combat = False

    if config.ITEM_SYSTEM_ENABLED:
      self.inventory = InventoryObs(inventory[0:config.INVENTORY_N_OBS],
                                    ItemState.State.attr_name_to_col["id"])
    else:
      assert inventory.size == 0

    if config.EXCHANGE_SYSTEM_ENABLED:
      self.market = BasicObs(market[0:config.MARKET_N_OBS],
                             ItemState.State.attr_name_to_col["id"])
    else:
      assert market.size == 0

    self._noop_action = 1 if config.PROVIDE_NOOP_ACTION_TARGET else 0

  # pylint: disable=method-cache-max-size-none
  @lru_cache(maxsize=None)
  def tile(self, r_delta, c_delta):
    '''Return the array object corresponding to a nearby tile

    Args:
        r_delta: row offset from current agent
        c_delta: col offset from current agent

    Returns:
        Vector corresponding to the specified tile
    '''
    agent = self.agent()
    center = self.config.PLAYER_VISION_RADIUS
    tile_dim = self.config.PLAYER_VISION_DIAMETER
    mat_map = self.tiles[:,2].reshape(tile_dim,tile_dim)
    new_row = agent.row + r_delta
    new_col = agent.col + c_delta
    if (0 <= new_row < self.config.MAP_SIZE) & \
       (0 <= new_col < self.config.MAP_SIZE):
      return TileState.parse_array([new_row, new_col, mat_map[center+r_delta,center+c_delta]])

    # return a dummy void tile at (inf, inf)
    return TileState.parse_array([np.inf, np.inf, material.Void.index])

  # pylint: disable=method-cache-max-size-none
  @lru_cache(maxsize=None)
  def entity(self, entity_id):
    rows = self.entities.values[self.entities.ids == entity_id]
    if rows.shape[0] == 0:
      return None
    return EntityState.parse_array(rows[0])

  # pylint: disable=method-cache-max-size-none
  @lru_cache(maxsize=None)
  def agent(self):
    return self.entity(self.agent_id)

  def clear_cache(self):
    # clear the cache, so that this object can be garbage collected
    self.agent.cache_clear()
    self.entity.cache_clear()
    self.tile.cache_clear()

  def get_empty_obs(self):
    gym_obs = {
      "CurrentTick": self.current_tick,
      "AgentId": self.agent_id,
      "Task": self.task_embedding,
      "Tile": None, # np.zeros((self.config.MAP_N_OBS, self.tiles.shape[1])),
      "Entity": np.zeros((self.config.PLAYER_N_OBS,
                          self.entities.values.shape[1]), dtype=np.int16)}
    if self.config.ITEM_SYSTEM_ENABLED:
      gym_obs["Inventory"] = np.zeros((self.config.INVENTORY_N_OBS,
                                       self.inventory.values.shape[1]), dtype=np.int16)
    if self.config.EXCHANGE_SYSTEM_ENABLED:
      gym_obs["Market"] = np.zeros((self.config.MARKET_N_OBS,
                                    self.market.values.shape[1]), dtype=np.int16)
    return gym_obs

  def to_gym(self):
    '''Convert the observation to a format that can be used by OpenAI Gym'''
    gym_obs = self.get_empty_obs()
    if self.dummy_obs:
      # return empty obs for the dead agents
      gym_obs['Tile'] = np.zeros((self.config.MAP_N_OBS, self.tiles.shape[1]), dtype=np.int16)
      if self.config.PROVIDE_ACTION_TARGETS:
        gym_obs["ActionTargets"] = self._make_action_targets()
      return gym_obs

    # NOTE: assume that all len(self.tiles) == self.config.MAP_N_OBS
    gym_obs['Tile'] = self.tiles
    gym_obs['Entity'][:self.entities.values.shape[0],:] = self.entities.values

    if self.config.ITEM_SYSTEM_ENABLED:
      gym_obs["Inventory"][:self.inventory.values.shape[0],:] = self.inventory.values

    if self.config.EXCHANGE_SYSTEM_ENABLED:
      gym_obs["Market"][:self.market.values.shape[0],:] = self.market.values

    if self.config.PROVIDE_ACTION_TARGETS:
      gym_obs["ActionTargets"] = self._make_action_targets()

    return gym_obs

  def _make_action_targets(self):
    masks = {}
    masks["Move"] = {
      "Direction": self._make_move_mask()
    }

    if self.config.COMBAT_SYSTEM_ENABLED:
      # Test below. see tests/core/test_observation_tile.py, test_action_target_consts()
      # assert len(action.Style.edges) == 3
      masks["Attack"] = {
        "Style": np.ones(3, dtype=np.int8),
        "Target": self._make_attack_mask()
      }

    if self.config.ITEM_SYSTEM_ENABLED:
      masks["Use"] = {
        "InventoryItem": self._make_use_mask()
      }
      masks["Give"] = {
        "InventoryItem": self._make_sell_mask(),
        "Target": self._make_give_target_mask()
      }
      masks["Destroy"] = {
        "InventoryItem": self._make_destroy_item_mask()
      }

    if self.config.EXCHANGE_SYSTEM_ENABLED:
      masks["Sell"] = {
        "InventoryItem": self._make_sell_mask(),
        "Price": np.ones(self.config.PRICE_N_OBS, dtype=np.int8)
      }
      masks["Buy"] = {
        "MarketItem": self._make_buy_mask()
      }
      masks["GiveGold"] = {
        "Price": self._make_give_gold_mask(),  # reusing Price
        "Target": self._make_give_gold_target_mask()
      }

    if self.config.COMMUNICATION_SYSTEM_ENABLED:
      masks["Comm"] = {
        "Token":np.ones(self.config.COMMUNICATION_NUM_TOKENS, dtype=np.int8)
      }

    return masks

  def _make_move_mask(self):
    if self.dummy_obs:
      mask = np.zeros(len(action.Direction.edges), dtype=np.int8)
      mask[-1] = 1  # for no-op
      return mask

    # pylint: disable=not-an-iterable
    return np.array([self.tile(*d.delta).material_id in material.Habitable.indices
                     for d in action.Direction.edges], dtype=np.int8)

  def _make_attack_mask(self):
    # NOTE: Currently, all attacks have the same range
    #   if we choose to make ranges different, the masks
    #   should be differently generated by attack styles
    assert self.config.COMBAT_MELEE_REACH == self.config.COMBAT_RANGE_REACH
    assert self.config.COMBAT_MELEE_REACH == self.config.COMBAT_MAGE_REACH
    assert self.config.COMBAT_RANGE_REACH == self.config.COMBAT_MAGE_REACH

    attack_mask = np.zeros(self.config.PLAYER_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      attack_mask[-1] = 1

    if self.dummy_obs:
      return attack_mask

    agent = self.agent()
    within_range = np.maximum( # calculating the l-inf dist
        np.abs(self.entities.values[:,EntityState.State.attr_name_to_col["row"]] - agent.row),
        np.abs(self.entities.values[:,EntityState.State.attr_name_to_col["col"]] - agent.col)
      ) <= self.config.COMBAT_MELEE_REACH

    immunity = self.config.COMBAT_SPAWN_IMMUNITY
    if agent.time_alive < immunity:
      # NOTE: CANNOT attack players during immunity, thus mask should set to 0
      no_spawn_immunity = ~(self.entities.ids > 0)  # ids > 0 equals entity.is_player
    else:
      no_spawn_immunity = np.ones(self.entities.len, dtype=bool)

    # allow friendly fire but no self shooting
    not_me = self.entities.ids != agent.id

    attack_mask[:self.entities.len] = within_range & not_me & no_spawn_immunity
    if sum(attack_mask[:self.entities.len]) > 0:
      # Mask the no-op option, since there should be at least one allowed move
      # NOTE: this will make agents always attack if there is a valid target
      attack_mask[-1] = 0

    return attack_mask

  def _make_use_mask(self):
    # empty inventory -- nothing to use
    use_mask = np.zeros(self.config.INVENTORY_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      use_mask[-1] = 1

    if not (self.config.ITEM_SYSTEM_ENABLED and self.inventory.len > 0)\
        or self.dummy_obs or self.agent_in_combat:
      return use_mask

    item_skill = self._item_skill()

    not_listed = self.inventory.values[:,ItemState.State.attr_name_to_col["listed_price"]] == 0
    item_type = self.inventory.values[:,ItemState.State.attr_name_to_col["type_id"]]
    item_level = self.inventory.values[:,ItemState.State.attr_name_to_col["level"]]

    # level limits are differently applied depending on item types
    type_flt = np.tile(np.array(list(item_skill.keys())), (self.inventory.len,1))
    level_flt = np.tile(np.array(list(item_skill.values())), (self.inventory.len,1))
    item_type = np.tile(np.transpose(np.atleast_2d(item_type)), (1,len(item_skill)))
    item_level = np.tile(np.transpose(np.atleast_2d(item_level)), (1,len(item_skill)))
    level_satisfied = np.any((item_type==type_flt) & (item_level<=level_flt), axis=1)

    use_mask[:self.inventory.len] = not_listed & level_satisfied
    return use_mask

  def _item_skill(self):
    agent = self.agent()

    # the minimum agent level is 1
    level = max(1, agent.melee_level, agent.range_level, agent.mage_level,
                agent.fishing_level, agent.herbalism_level, agent.prospecting_level,
                agent.carving_level, agent.alchemy_level)
    return {
      item_system.Hat.ITEM_TYPE_ID: level,
      item_system.Top.ITEM_TYPE_ID: level,
      item_system.Bottom.ITEM_TYPE_ID: level,
      item_system.Spear.ITEM_TYPE_ID: agent.melee_level,
      item_system.Bow.ITEM_TYPE_ID: agent.range_level,
      item_system.Wand.ITEM_TYPE_ID: agent.mage_level,
      item_system.Rod.ITEM_TYPE_ID: agent.fishing_level,
      item_system.Gloves.ITEM_TYPE_ID: agent.herbalism_level,
      item_system.Pickaxe.ITEM_TYPE_ID: agent.prospecting_level,
      item_system.Axe.ITEM_TYPE_ID: agent.carving_level,
      item_system.Chisel.ITEM_TYPE_ID: agent.alchemy_level,
      item_system.Whetstone.ITEM_TYPE_ID: agent.melee_level,
      item_system.Arrow.ITEM_TYPE_ID: agent.range_level,
      item_system.Runes.ITEM_TYPE_ID: agent.mage_level,
      item_system.Ration.ITEM_TYPE_ID: level,
      item_system.Potion.ITEM_TYPE_ID: level
    }

  def _make_destroy_item_mask(self):
    destroy_mask = np.zeros(self.config.INVENTORY_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      destroy_mask[-1] = 1

    # empty inventory -- nothing to destroy
    if not (self.config.ITEM_SYSTEM_ENABLED and self.inventory.len > 0)\
        or self.dummy_obs or self.agent_in_combat:
      return destroy_mask

    # not equipped items in the inventory can be destroyed
    not_equipped = self.inventory.values[:,ItemState.State.attr_name_to_col["equipped"]] == 0

    destroy_mask[:self.inventory.len] = not_equipped
    return destroy_mask

  def _make_give_target_mask(self):
    give_mask = np.zeros(self.config.PLAYER_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      give_mask[-1] = 1

    if not self.config.ITEM_SYSTEM_ENABLED or self.dummy_obs or self.agent_in_combat\
       or self.inventory.len == 0:
      return give_mask

    agent = self.agent()
    entities_pos = self.entities.values[:,[EntityState.State.attr_name_to_col["row"],
                                           EntityState.State.attr_name_to_col["col"]]]
    same_tile = utils.linf(entities_pos, (agent.row, agent.col)) == 0
    not_me = self.entities.ids != self.agent_id
    player = (self.entities.values[:,EntityState.State.attr_name_to_col["npc_type"]] == 0)

    give_mask[:self.entities.len] = same_tile & player & not_me
    return give_mask

  def _make_give_gold_target_mask(self):
    give_mask = np.zeros(self.config.PLAYER_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      give_mask[-1] = 1

    if not self.config.EXCHANGE_SYSTEM_ENABLED or self.dummy_obs or self.agent_in_combat\
       or int(self.agent().gold) == 0:
      return give_mask

    agent = self.agent()
    entities_pos = self.entities.values[:,[EntityState.State.attr_name_to_col["row"],
                                           EntityState.State.attr_name_to_col["col"]]]
    same_tile = utils.linf(entities_pos, (agent.row, agent.col)) == 0
    not_me = self.entities.ids != self.agent_id
    player = (self.entities.values[:,EntityState.State.attr_name_to_col["npc_type"]] == 0)

    give_mask[:self.entities.len] = same_tile & player & not_me
    return give_mask

  def _make_give_gold_mask(self):
    mask = np.zeros(self.config.PRICE_N_OBS, dtype=np.int8)
    mask[0] = 1  # To avoid all-0 masks. If the agent has no gold, this action will be ignored.
    if self.dummy_obs or self.agent_in_combat:
      return mask

    gold = int(self.agent().gold)
    mask[:gold] = 1 # NOTE that action.Price starts from Discrete_1
    return mask

  def _make_sell_mask(self):
    sell_mask = np.zeros(self.config.INVENTORY_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      sell_mask[-1] = 1

    # empty inventory -- nothing to sell
    if not (self.config.EXCHANGE_SYSTEM_ENABLED and self.inventory.len > 0) \
      or self.dummy_obs or self.agent_in_combat:
      return sell_mask

    not_equipped = self.inventory.values[:,ItemState.State.attr_name_to_col["equipped"]] == 0
    not_listed = self.inventory.values[:,ItemState.State.attr_name_to_col["listed_price"]] == 0

    sell_mask[:self.inventory.len] = not_equipped & not_listed
    return sell_mask

  def _make_buy_mask(self):
    buy_mask = np.zeros(self.config.MARKET_N_OBS + self._noop_action, dtype=np.int8)
    if self.config.PROVIDE_NOOP_ACTION_TARGET:
      buy_mask[-1] = 1

    if not self.config.EXCHANGE_SYSTEM_ENABLED or self.dummy_obs or self.agent_in_combat \
       or self.market.len == 0:
      return buy_mask

    agent = self.agent()
    market_items = self.market.values
    not_mine = market_items[:,ItemState.State.attr_name_to_col["owner_id"]] != self.agent_id

    # if the inventory is full, one can only buy existing ammo stack
    #   otherwise, one can buy anything owned by other, having enough money
    if self.inventory.len >= self.config.ITEM_INVENTORY_CAPACITY:
      exist_ammo_listings = self._existing_ammo_listings()
      if not np.any(exist_ammo_listings):
        return buy_mask
      not_mine &= exist_ammo_listings

    enough_gold = market_items[:,ItemState.State.attr_name_to_col["listed_price"]] <= agent.gold
    buy_mask[:self.market.len] = not_mine & enough_gold
    return buy_mask

  def _existing_ammo_listings(self):
    sig_col = (ItemState.State.attr_name_to_col["type_id"],
               ItemState.State.attr_name_to_col["level"])
    ammo_id = [ammo.ITEM_TYPE_ID for ammo in
              [item_system.Whetstone, item_system.Arrow, item_system.Runes]]

    # search ammo stack from the inventory
    type_flt = np.tile(np.array(ammo_id), (self.inventory.len,1))
    item_type = np.tile(
      np.transpose(np.atleast_2d(self.inventory.values[:,sig_col[0]])),
      (1, len(ammo_id)))
    exist_ammo = self.inventory.values[np.any(item_type == type_flt, axis=1)]

    # self does not have ammo
    if exist_ammo.shape[0] == 0:
      return np.zeros(self.market.len, dtype=bool)

    # search the existing ammo stack from the market that's not mine
    type_flt = np.tile(np.array(exist_ammo[:,sig_col[0]]), (self.market.len,1))
    level_flt = np.tile(np.array(exist_ammo[:,sig_col[1]]), (self.market.len,1))
    item_type = np.tile(np.transpose(np.atleast_2d(self.market.values[:,sig_col[0]])),
                        (1, exist_ammo.shape[0]))
    item_level = np.tile(np.transpose(np.atleast_2d(self.market.values[:,sig_col[1]])),
                         (1, exist_ammo.shape[0]))
    exist_ammo_listings = np.any((item_type==type_flt) & (item_level==level_flt), axis=1)

    not_mine = self.market.values[:,ItemState.State.attr_name_to_col["owner_id"]] != self.agent_id

    return exist_ammo_listings & not_mine



# pylint: disable=all

import numpy as np

from signal import signal, SIGINT
import json
import os
import sys
import time
import threading

from twisted.internet import reactor
from twisted.python import log
from twisted.web.server import Site
from twisted.web.static import File

from autobahn.twisted.websocket import WebSocketServerFactory, \
    WebSocketServerProtocol
from autobahn.twisted.resource import WebSocketResource

from .render_utils import np_encoder

class GodswordServerProtocol(WebSocketServerProtocol):
  def __init__(self):
    super().__init__()
    print("Created a server")
    self.frame = 0

    #"connected" is already used by WSSP
    self.sent_environment = False
    self.isConnected      = False

    self.pos = [0, 0]
    self.cmd = None

  def onOpen(self):
    print("Opened connection to server")

  def onClose(self, wasClean, code=None, reason=None):
    self.isConnected = False
    print('Connection closed')

  def connectionMade(self):
    super().connectionMade()
    self.factory.clientConnectionMade(self)

  def connectionLost(self, reason):
    super().connectionLost(reason)
    self.factory.clientConnectionLost(self)
    self.sent_environment = False

  #Not used without player interaction
  def onMessage(self, packet, isBinary):
    print("Server packet", packet)
    packet    = packet.decode()
    _, packet = packet.split(';') #Strip headeer
    r, c, cmd = packet.split(' ') #Split camera coords
    if len(cmd) == 0 or cmd == '\t':
        cmd = None

    self.pos = [int(r), int(c)]
    self.cmd = cmd

    self.isConnected = True

  def onConnect(self, request):
    print("WebSocket connection request: {}".format(request))
    realm = self.factory.realm
    self.realm = realm
    self.frame += 1

  def serverPacket(self):
    data = self.realm.packet
    return data

  def sendUpdate(self, data):
    packet               = {}
    packet['resource']   = data['resource']
    packet['player']     = data['player']
    packet['npc']        = data['npc']
    packet['pos']        = data['pos']
    packet['wilderness'] = data['wilderness']
    packet['market']     = data['market']

    print('Is Connected? : {}'.format(self.isConnected))
    if not self.sent_environment:
      packet['map']    = data['environment']
      packet['border'] = data['border']
      packet['size']   = data['size']
      self.sent_environment=True

    if 'overlay' in data:
      packet['overlay'] = data['overlay']
      print('SENDING OVERLAY: ', len(packet['overlay']))

    packet = json.dumps(packet, default=np_encoder).encode('utf8')
    self.sendMessage(packet, False)


class WSServerFactory(WebSocketServerFactory):
  def __init__(self, ip, realm):
    super().__init__(ip)
    self.realm = realm
    self.time = time.time()
    self.clients = []

    self.pos = [0, 0]
    self.cmd = None
    self.tickRate = 0.6
    self.tick = 0

  def update(self, packet):
    self.tick += 1
    uptime = np.round(self.tickRate*self.tick, 1)
    delta = time.time() - self.time
    print('Wall Clock: ', str(delta)[:5], 'Uptime: ', uptime, ', Tick: ', self.tick)
    delta = self.tickRate - delta
    if delta > 0:
      time.sleep(delta)
    self.time = time.time()

    for client in self.clients:
      client.sendUpdate(packet)
      if client.pos is not None:
        self.pos = client.pos
        self.cmd = client.cmd

    return self.pos, self.cmd

  def clientConnectionMade(self, client):
    self.clients.append(client)

  def clientConnectionLost(self, client):
    self.clients.remove(client)

class Application:
  def __init__(self, realm):
    signal(SIGINT, self.kill)
    log.startLogging(sys.stdout)

    port = 8080
    self.factory          = WSServerFactory(u'ws://localhost:{}'.format(port), realm)
    self.factory.protocol = GodswordServerProtocol
    resource              = WebSocketResource(self.factory)

    root = File(".")
    root.putChild(b"ws", resource)
    site = Site(root)

    reactor.listenTCP(port, site)

    def run():
      reactor.run(installSignalHandlers=0)

    threading.Thread(target=run).start()

  def update(self, packet):
    return self.factory.update(packet)

  def kill(*args):
    print("Killed by user")
    reactor.stop()
    os._exit(0)


from nmmo.entity import entity
from nmmo.core import action as Action
from nmmo.systems import combat, droptable
from nmmo.systems.ai import policy
from nmmo.systems import item as Item
from nmmo.systems import skill
from nmmo.systems.inventory import EquipmentSlot
from nmmo.lib.log import EventCode

class Equipment:
  def __init__(self, total,
    melee_attack, range_attack, mage_attack,
    melee_defense, range_defense, mage_defense):

    self.level         = total
    self.ammunition    = EquipmentSlot()

    self.melee_attack  = melee_attack
    self.range_attack  = range_attack
    self.mage_attack   = mage_attack
    self.melee_defense = melee_defense
    self.range_defense = range_defense
    self.mage_defense  = mage_defense

  def total(self, getter):
    return getter(self)

  # pylint: disable=R0801
  # Similar lines here and in inventory.py
  @property
  def packet(self):
    packet = {}

    packet['item_level']    = self.total
    packet['melee_attack']  = self.melee_attack
    packet['range_attack']  = self.range_attack
    packet['mage_attack']   = self.mage_attack
    packet['melee_defense'] = self.melee_defense
    packet['range_defense'] = self.range_defense
    packet['mage_defense']  = self.mage_defense

    return packet


# pylint: disable=no-member
class NPC(entity.Entity):
  def __init__(self, realm, pos, iden, name, npc_type):
    super().__init__(realm, pos, iden, name)
    self.skills = skill.Combat(realm, self)
    self.realm = realm
    self.last_action = None
    self.droptable = None
    self.spawn_danger = None
    self.equipment = None
    self.npc_type.update(npc_type)

  def update(self, realm, actions):
    super().update(realm, actions)

    if not self.alive:
      return

    self.resources.health.increment(1)
    self.last_action = actions

  # Returns True if the entity is alive
  def receive_damage(self, source, dmg):
    if super().receive_damage(source, dmg):
      return True

    # run the next lines if the npc is killed
    # source receive gold & items in the droptable
    # pylint: disable=no-member
    if self.gold.val > 0:
      source.gold.increment(self.gold.val)
      self.realm.event_log.record(EventCode.EARN_GOLD, source, amount=self.gold.val)
      self.gold.update(0)

    for item in self.droptable.roll(self.realm, self.attack_level):
      if source.is_player and source.inventory.space:
        # inventory.receive() returns True if the item is received
        # if source doesn't have space, inventory.receive() destroys the item
        if source.inventory.receive(item):
          self.realm.event_log.record(EventCode.LOOT_ITEM, source, item=item)
      else:
        item.destroy()

    return False

  # NOTE: passing np_random here is a hack
  #   Ideally, it should be passed to __init__ and also used in action generation
  @staticmethod
  def spawn(realm, pos, iden, np_random):
    config = realm.config

    # check the position
    if realm.map.tiles[pos].impassible:
      return None

    # Select AI Policy
    danger = combat.danger(config, pos)
    if danger >= config.NPC_SPAWN_AGGRESSIVE:
      ent = Aggressive(realm, pos, iden)
    elif danger >= config.NPC_SPAWN_NEUTRAL:
      ent = PassiveAggressive(realm, pos, iden)
    elif danger >= config.NPC_SPAWN_PASSIVE:
      ent = Passive(realm, pos, iden)
    else:
      return None

    ent.spawn_danger = danger

    # Select combat focus
    style = np_random.integers(0,3)
    if style == 0:
      style = Action.Melee
    elif style == 1:
      style = Action.Range
    else:
      style = Action.Mage
    ent.skills.style = style

    # Compute level
    level = 0
    if config.PROGRESSION_SYSTEM_ENABLED:
      level_min = config.NPC_LEVEL_MIN
      level_max = config.NPC_LEVEL_MAX
      level     = int(danger * (level_max - level_min) + level_min)

      # Set skill levels
      if style == Action.Melee:
        ent.skills.melee.set_experience_by_level(level)
      elif style == Action.Range:
        ent.skills.range.set_experience_by_level(level)
      elif style == Action.Mage:
        ent.skills.mage.set_experience_by_level(level)

    # Gold
    if config.EXCHANGE_SYSTEM_ENABLED:
      # pylint: disable=no-member
      ent.gold.update(level)

    ent.droptable = droptable.Standard()

    # Equipment to instantiate
    if config.EQUIPMENT_SYSTEM_ENABLED:
      lvl     = level - np_random.random()
      ilvl    = int(5 * lvl)

      offense = int(config.NPC_BASE_DAMAGE + lvl*config.NPC_LEVEL_DAMAGE)
      defense = int(config.NPC_BASE_DEFENSE + lvl*config.NPC_LEVEL_DEFENSE)

      ent.equipment = Equipment(ilvl, offense, offense, offense, defense, defense, defense)

      armor =  [Item.Hat, Item.Top, Item.Bottom]
      ent.droptable.add(np_random.choice(armor))

    if config.PROFESSION_SYSTEM_ENABLED:
      tools =  [Item.Rod, Item.Gloves, Item.Pickaxe, Item.Axe, Item.Chisel]
      ent.droptable.add(np_random.choice(tools))

    return ent

  def packet(self):
    data = super().packet()

    data['skills']   = self.skills.packet()
    data['resource'] = { 'health': {
      'val': self.resources.health.val, 'max': self.config.PLAYER_BASE_HEALTH } }

    return data

  @property
  def is_npc(self) -> bool:
    return True

class Passive(NPC):
  def __init__(self, realm, pos, iden):
    super().__init__(realm, pos, iden, 'Passive', 1)

  def decide(self, realm):
    return policy.passive(realm, self)

class PassiveAggressive(NPC):
  def __init__(self, realm, pos, iden):
    super().__init__(realm, pos, iden, 'Neutral', 2)

  def decide(self, realm):
    return policy.neutral(realm, self)

class Aggressive(NPC):
  def __init__(self, realm, pos, iden):
    super().__init__(realm, pos, iden, 'Hostile', 3)

  def decide(self, realm):
    return policy.hostile(realm, self)



from __future__ import annotations
from typing import Dict, Union, Iterable, TYPE_CHECKING
from collections import OrderedDict
from collections.abc import Set, Sequence
import weakref

if TYPE_CHECKING:
  from nmmo.task.game_state import GameState, GroupView

class Group(Sequence, Set):
  ''' An immutable, ordered, unique group of agents involved in a task
  '''
  def __init__(self,
               agents: Union(Iterable[int], int),
               name: str=None):

    if isinstance(agents, int):
      agents = (agents,)
    assert len(agents) > 0, "Team must have at least one agent"
    self.name = name if name else f"Agent({','.join([str(e) for e in agents])})"
    # Remove duplicates
    self._agents = tuple(OrderedDict.fromkeys(sorted(agents)).keys())
    if not isinstance(self._agents,tuple):
      self._agents = (self._agents,)

    self._sd: GroupView = None
    self._gs: GameState = None

    self._hash = hash(self._agents)

  @property
  def agents(self):
    return self._agents

  def union(self, o: Group):
    return Group(self._agents + o.agents)

  def intersection(self, o: Group):
    return Group(set(self._agents).intersection(set(o.agents)))

  def __eq__(self, o):
    return self._agents == o

  def __len__(self):
    return len(self._agents)

  def __hash__(self):
    return self._hash

  def __getitem__(self, key):
    if len(self) == 1 and key == 0:
      return self
    return Group((self._agents[key],), f"{self.name}.{key}")

  def __contains__(self, key):
    if isinstance(key, int):
      return key in self.agents
    return Sequence.__contains__(self, key)

  def __str__(self) -> str:
    return str(self._agents)

  def __int__(self) -> int:
    assert len(self._agents) == 1, "Group is not a singleton"
    return int(self._agents[0])

  def __copy__(self):
    return self
  def __deepcopy__(self, memo):
    return Group(self.agents, self.name)

  def description(self) -> Dict:
    return {
      "type": "Group",
      "name": self.name,
      "agents": self._agents
    }

  def clear_prev_state(self) -> None:
    if self._gs is not None:
      self._gs.clear_cache()  # prevent memory leak
      self._gs = None
    if self._sd is not None:
      weakref.ref(self._sd)  # prevent memory leak
      self._sd = None

  def update(self, gs: GameState) -> None:
    self.clear_prev_state()
    self._gs = gs
    self._sd = gs.get_subject_view(self)

  def __getattr__(self, attr):
    return self._sd.__getattribute__(attr)

def union(*groups: Group) -> Group:
  """ Performs a big union over groups
  """
  agents = []
  for group in groups:
    for agent in group.agents:
      agents.append(agent)
  return Group(agents)

def complement(group: Group, universe: Group) -> Group:
  """ Returns the complement of group in universe
  """
  agents = []
  for agent in universe.agents:
    if not agent in group:
      agents.append(agent)
  return Group(agents)



from types import SimpleNamespace

from nmmo.datastore.serialized import SerializedState
from nmmo.lib import material

# pylint: disable=no-member,protected-access
TileState = SerializedState.subclass(
  "Tile", [
    "row",
    "col",
    "material_id",
  ])

TileState.Limits = lambda config: {
  "row": (0, config.MAP_SIZE-1),
  "col": (0, config.MAP_SIZE-1),
  "material_id": (0, config.MAP_N_TILE),
}

TileState.Query = SimpleNamespace(
  window=lambda ds, r, c, radius: ds.table("Tile").window(
    TileState.State.attr_name_to_col["row"],
    TileState.State.attr_name_to_col["col"],
    r, c, radius),
  get_map=lambda ds, map_size:
    ds.table("Tile")._data[1:(map_size*map_size+1)]
                    .reshape((map_size,map_size,len(TileState.State.attr_name_to_col)))
)

class Tile(TileState):
  def __init__(self, realm, r, c, np_random):
    super().__init__(realm.datastore, TileState.Limits(realm.config))
    self.realm = realm
    self.config = realm.config
    self._np_random = np_random

    self.row.update(r)
    self.col.update(c)

    self.state = None
    self.material = None
    self.depleted = False
    self.tex = None

    self.entities = {}

  @property
  def repr(self):
    return ((self.row.val, self.col.val))

  @property
  def pos(self):
    return self.row.val, self.col.val

  @property
  def habitable(self):
    return self.material in material.Habitable

  @property
  def impassible(self):
    return self.material in material.Impassible

  @property
  def void(self):
    return self.material == material.Void

  def reset(self, mat, config, np_random):
    self._np_random = np_random # reset the RNG
    self.state = mat(config)
    self.material = mat(config)
    self.material_id.update(self.state.index)

    self.depleted = False
    self.tex = self.material.tex

    self.entities = {}

  def add_entity(self, ent):
    assert ent.ent_id not in self.entities
    self.entities[ent.ent_id] = ent

  def remove_entity(self, ent_id):
    assert ent_id in self.entities
    del self.entities[ent_id]

  def step(self):
    if not self.depleted or self._np_random.random() > self.material.respawn:
      return

    self.depleted = False
    self.state = self.material
    self.material_id.update(self.state.index)

  def harvest(self, deplete):
    assert not self.depleted, f'{self.state} is depleted'
    assert self.state in material.Harvestable, f'{self.state} not harvestable'

    if deplete:
      self.depleted = True
      self.state = self.material.deplete(self.config)
      self.material_id.update(self.state.index)

    return self.material.harvest()



#pylint: disable=protected-access, invalid-name

import numpy as np

import nmmo
from nmmo.systems.ai import move, utils

def update(entity):
  '''Update validity of tracked entities'''
  if not utils.validTarget(entity, entity.attacker, entity.vision):
    entity.attacker = None
  if not utils.validTarget(entity, entity.target, entity.vision):
    entity.target = None
  if not utils.validTarget(entity, entity.closest, entity.vision):
    entity.closest = None

  if entity.__class__.__name__ != 'Player':
    return

  if not utils.validResource(entity, entity.food, entity.vision):
    entity.food = None
  if not utils.validResource(entity, entity.water, entity.vision):
    entity.water = None


def pathfind(realm, actions, entity, target):
  # TODO: do not access realm._np_random directly. ALSO see below for all other uses
  actions[nmmo.action.Move] = {
    nmmo.action.Direction: move.pathfind(realm.map, entity, target, realm._np_random)}


def explore(realm, actions, entity):
  sz = realm.config.TERRAIN_SIZE
  r, c = entity.pos

  spawnR, spawnC = entity.spawnPos
  centR, centC = sz//2, sz//2

  vR, vC = centR-spawnR, centC-spawnC

  mmag = max(abs(vR), abs(vC))
  rr = r + int(np.round(entity.vision*vR/mmag))
  cc = c + int(np.round(entity.vision*vC/mmag))

  tile = realm.map.tiles[rr, cc]
  pathfind(realm, actions, entity, tile)


def meander(realm, actions, entity):
  actions[nmmo.action.Move] = {
    nmmo.action.Direction: move.habitable(realm.map, entity, realm._np_random)}

def evade(realm, actions, entity):
  actions[nmmo.action.Move] = {
    nmmo.action.Direction: move.antipathfind(realm.map, entity, entity.attacker,
                                                realm._np_random)}

def hunt(realm, actions, entity):
  # Move args
  distance = utils.lInfty(entity.pos, entity.target.pos)

  if distance > 1:
    actions[nmmo.action.Move] = {nmmo.action.Direction: move.pathfind(realm.map,
                                                                      entity,
                                                                      entity.target,
                                                                      realm._np_random)}
  elif distance == 0:
    actions[nmmo.action.Move] = {
        nmmo.action.Direction: move.habitable(realm.map, entity, realm._np_random)}

  attack(realm, actions, entity)

def attack(realm, actions, entity):
  distance = utils.lInfty(entity.pos, entity.target.pos)
  if distance > entity.skills.style.attack_range(realm.config):
    return

  actions[nmmo.action.Attack] = {
    nmmo.action.Style: entity.skills.style,
    nmmo.action.Target: entity.target}



# CHECK ME: Should these be fixed as well?
# pylint: disable=no-method-argument,unused-argument,no-self-argument,no-member

from enum import Enum, auto
import numpy as np
from nmmo.core.observation import Observation

from nmmo.lib import utils
from nmmo.lib.utils import staticproperty
from nmmo.systems.item import Stack
from nmmo.lib.log import EventCode


class NodeType(Enum):
  #Tree edges
  STATIC = auto()    #Traverses all edges without decisions
  SELECTION = auto() #Picks an edge to follow

  #Executable actions
  ACTION    = auto() #No arguments
  CONSTANT  = auto() #Constant argument
  VARIABLE  = auto() #Variable argument

class Node(metaclass=utils.IterableNameComparable):
  @classmethod
  def init(cls, config):
    # noop_action is used in some of the N() methods
    cls.noop_action = 1 if config.PROVIDE_NOOP_ACTION_TARGET else 0

  @staticproperty
  def edges():
    return []

  #Fill these in
  @staticproperty
  def priority():
    return None

  @staticproperty
  def type():
    return None

  @staticproperty
  def leaf():
    return False

  @classmethod
  def N(cls, config):
    return len(cls.edges)

  def deserialize(realm, entity, index, obs: Observation):
    return index

class Fixed:
  pass

#ActionRoot
class Action(Node):
  nodeType = NodeType.SELECTION
  hooked   = False

  @classmethod
  def init(cls, config):
    # Sets up serialization domain
    if Action.hooked:
      return

    Action.hooked = True

  #Called upon module import (see bottom of file)
  #Sets up serialization domain
  def hook(config):
    idx = 0
    arguments = []
    for action in Action.edges(config):
      action.init(config)
      for args in action.edges: # pylint: disable=not-an-iterable
        args.init(config)
        if not 'edges' in args.__dict__:
          continue
        for arg in args.edges:
          arguments.append(arg)
          arg.serial = tuple([idx])
          arg.idx = idx
          idx += 1
    Action.arguments = arguments

  @staticproperty
  def n():
    return len(Action.arguments)

  # pylint: disable=invalid-overridden-method
  @classmethod
  def edges(cls, config):
    '''List of valid actions'''
    edges = [Move]
    if config.COMBAT_SYSTEM_ENABLED:
      edges.append(Attack)
    if config.ITEM_SYSTEM_ENABLED:
      edges += [Use, Give, Destroy]
    if config.EXCHANGE_SYSTEM_ENABLED:
      edges += [Buy, Sell, GiveGold]
    if config.COMMUNICATION_SYSTEM_ENABLED:
      edges.append(Comm)
    return edges

class Move(Node):
  priority = 60
  nodeType = NodeType.SELECTION
  def call(realm, entity, direction):
    if direction is None:
      return

    assert entity.alive, "Dead entity cannot act"
    assert realm.map.is_valid_pos(*entity.pos), "Invalid entity position"

    r, c  = entity.pos
    ent_id = entity.ent_id
    entity.history.last_pos = (r, c)
    r_delta, c_delta = direction.delta
    r_new, c_new = r+r_delta, c+c_delta

    if not realm.map.is_valid_pos(r_new, c_new) or \
       realm.map.tiles[r_new, c_new].impassible:
      return

    if entity.status.freeze > 0:
      return

    entity.row.update(r_new)
    entity.col.update(c_new)

    realm.map.tiles[r, c].remove_entity(ent_id)
    realm.map.tiles[r_new, c_new].add_entity(entity)

    # exploration record keeping. moved from entity.py, History.update()
    progress_to_center = realm.map.dist_border_center -\
      utils.linf_single(realm.map.center_coord, (r_new, c_new))
    if progress_to_center > entity.history.exploration:
      entity.history.exploration = progress_to_center
      if entity.is_player:
        realm.event_log.record(EventCode.GO_FARTHEST, entity,
                               distance=progress_to_center)

    # CHECK ME: material.Impassible includes void, so this line is not reachable
    #   Does this belong to Entity/Player.update()?
    if realm.map.tiles[r_new, c_new].void:
      entity.receive_damage(None, entity.resources.health.val)

  @staticproperty
  def edges():
    return [Direction]

  @staticproperty
  def leaf():
    return True

  def enabled(config):
    return True

class Direction(Node):
  argType = Fixed

  @staticproperty
  def edges():
    return [North, South, East, West, Stay]

  def deserialize(realm, entity, index, obs: Observation):
    return deserialize_fixed_arg(Direction, index)

# a quick helper function
def deserialize_fixed_arg(arg, index):
  if isinstance(index, (int, np.int64)):
    if index < 0:
      return None # so that the action will be discarded
    val = min(index, len(arg.edges)-1)
    return arg.edges[val]

  # if index is not int, it's probably already deserialized
  if index not in arg.edges:
    return None # so that the action will be discarded
  return index

class North(Node):
  delta = (-1, 0)

class South(Node):
  delta = (1, 0)

class East(Node):
  delta = (0, 1)

class West(Node):
  delta = (0, -1)

class Stay(Node):
  delta = (0, 0)

class Attack(Node):
  priority = 50
  nodeType = NodeType.SELECTION
  @staticproperty
  def n():
    return 3

  @staticproperty
  def edges():
    return [Style, Target]

  @staticproperty
  def leaf():
    return True

  def enabled(config):
    return config.COMBAT_SYSTEM_ENABLED

  def in_range(entity, stim, config, N):
    R, C = stim.shape
    R, C = R//2, C//2

    rets = set([entity])
    for r in range(R-N, R+N+1):
      for c in range(C-N, C+N+1):
        for e in stim[r, c].entities.values():
          rets.add(e)

    rets = list(rets)
    return rets

  def call(realm, entity, style, target):
    if style is None or target is None:
      return None

    assert entity.alive, "Dead entity cannot act"

    config = realm.config
    if entity.is_player and not config.COMBAT_SYSTEM_ENABLED:
      return None

    # Testing a spawn immunity against old agents to avoid spawn camping
    immunity = config.COMBAT_SPAWN_IMMUNITY
    if entity.is_player and target.is_player and \
      target.history.time_alive < immunity:
      return None

    #Check if self targeted
    if entity.ent_id == target.ent_id:
      return None

    #Can't attack out of range
    if utils.linf_single(entity.pos, target.pos) > style.attack_range(config):
      return None

    #Execute attack
    entity.history.attack = {}
    entity.history.attack['target'] = target.ent_id
    entity.history.attack['style'] = style.__name__
    target.attacker = entity
    target.attacker_id.update(entity.ent_id)

    from nmmo.systems import combat
    dmg = combat.attack(realm, entity, target, style.skill)

    if style.freeze and dmg > 0:
      target.status.freeze.update(config.COMBAT_FREEZE_TIME)

    # record the combat tick for both entities
    # players and npcs both have latest_combat_tick in EntityState
    for ent in [entity, target]:
      ent.latest_combat_tick.update(realm.tick + 1) # because the tick is about to increment

    return dmg

class Style(Node):
  argType = Fixed
  @staticproperty
  def edges():
    return [Melee, Range, Mage]

  def deserialize(realm, entity, index, obs: Observation):
    return deserialize_fixed_arg(Style, index)

class Target(Node):
  argType = None

  @classmethod
  def N(cls, config):
    return config.PLAYER_N_OBS + cls.noop_action

  def deserialize(realm, entity, index: int, obs: Observation):
    if index >= len(obs.entities.ids):
      return None
    return realm.entity_or_none(obs.entities.ids[index])

class Melee(Node):
  nodeType = NodeType.ACTION
  freeze=False

  def attack_range(config):
    return config.COMBAT_MELEE_REACH

  def skill(entity):
    return entity.skills.melee

class Range(Node):
  nodeType = NodeType.ACTION
  freeze=False

  def attack_range(config):
    return config.COMBAT_RANGE_REACH

  def skill(entity):
    return entity.skills.range

class Mage(Node):
  nodeType = NodeType.ACTION
  freeze=False

  def attack_range(config):
    return config.COMBAT_MAGE_REACH

  def skill(entity):
    return entity.skills.mage

class InventoryItem(Node):
  argType  = None

  @classmethod
  def N(cls, config):
    return config.INVENTORY_N_OBS + cls.noop_action

  def deserialize(realm, entity, index: int, obs: Observation):
    if index >= len(obs.inventory.ids):
      return None
    return realm.items.get(obs.inventory.ids[index])

class Use(Node):
  priority = 10

  @staticproperty
  def edges():
    return [InventoryItem]

  def enabled(config):
    return config.ITEM_SYSTEM_ENABLED

  def call(realm, entity, item):
    if item is None or item.owner_id.val != entity.ent_id:
      return

    assert entity.alive, "Dead entity cannot act"
    assert entity.is_player, "Npcs cannot use an item"
    assert item.quantity.val > 0, "Item quantity cannot be 0" # indicates item leak

    if not realm.config.ITEM_SYSTEM_ENABLED:
      return

    if item not in entity.inventory:
      return

    if entity.in_combat: # player cannot use item during combat
      return

    # cannot use listed items or items that have higher level
    if item.listed_price.val > 0 or item.level_gt(entity):
      return

    item.use(entity)

class Destroy(Node):
  priority = 40

  @staticproperty
  def edges():
    return [InventoryItem]

  def enabled(config):
    return config.ITEM_SYSTEM_ENABLED

  def call(realm, entity, item):
    if item is None or item.owner_id.val != entity.ent_id:
      return

    assert entity.alive, "Dead entity cannot act"
    assert entity.is_player, "Npcs cannot destroy an item"
    assert item.quantity.val > 0, "Item quantity cannot be 0" # indicates item leak

    if not realm.config.ITEM_SYSTEM_ENABLED:
      return

    if item not in entity.inventory:
      return

    if item.equipped.val: # cannot destroy equipped item
      return

    if entity.in_combat: # player cannot destroy item during combat
      return

    item.destroy()

    realm.event_log.record(EventCode.DESTROY_ITEM, entity)

class Give(Node):
  priority = 30

  @staticproperty
  def edges():
    return [InventoryItem, Target]

  def enabled(config):
    return config.ITEM_SYSTEM_ENABLED

  def call(realm, entity, item, target):
    if item is None or item.owner_id.val != entity.ent_id or target is None:
      return

    assert entity.alive, "Dead entity cannot act"
    assert entity.is_player, "Npcs cannot give an item"
    assert item.quantity.val > 0, "Item quantity cannot be 0" # indicates item leak

    config = realm.config
    if not config.ITEM_SYSTEM_ENABLED:
      return

    if not (target.is_player and target.alive):
      return

    if item not in entity.inventory:
      return

    # cannot give the equipped or listed item
    if item.equipped.val or item.listed_price.val:
      return

    if entity.in_combat: # player cannot give item during combat
      return

    if not (config.ITEM_ALLOW_GIFT and
            entity.ent_id != target.ent_id and                      # but not self
            target.is_player and
            entity.pos == target.pos):               # the same tile
      return

    if not target.inventory.space:
      # receiver inventory is full - see if it has an ammo stack with the same sig
      if isinstance(item, Stack):
        if not target.inventory.has_stack(item.signature):
          # no ammo stack with the same signature, so cannot give
          return
      else: # no space, and item is not ammo stack, so cannot give
        return

    entity.inventory.remove(item)
    target.inventory.receive(item)

    realm.event_log.record(EventCode.GIVE_ITEM, entity)

class GiveGold(Node):
  priority = 30

  @staticproperty
  def edges():
    # CHECK ME: for now using Price to indicate the gold amount to give
    return [Price, Target]

  def enabled(config):
    return config.EXCHANGE_SYSTEM_ENABLED

  def call(realm, entity, amount, target):
    if amount is None or target is None:
      return

    assert entity.alive, "Dead entity cannot act"
    assert entity.is_player, "Npcs cannot give gold"

    config = realm.config
    if not config.EXCHANGE_SYSTEM_ENABLED:
      return

    if not (target.is_player and target.alive):
      return

    if entity.in_combat: # player cannot give gold during combat
      return

    if not (config.ITEM_ALLOW_GIFT and
            entity.ent_id != target.ent_id and                      # but not self
            target.is_player and
            entity.pos == target.pos):                              # the same tile
      return

    if not isinstance(amount, int):
      amount = amount.val

    if amount > entity.gold.val: # no gold to give
      return

    entity.gold.decrement(amount)
    target.gold.increment(amount)

    realm.event_log.record(EventCode.GIVE_GOLD, entity)

class MarketItem(Node):
  argType  = None

  @classmethod
  def N(cls, config):
    return config.MARKET_N_OBS + cls.noop_action

  def deserialize(realm, entity, index: int, obs: Observation):
    if index >= len(obs.market.ids):
      return None

    return realm.items.get(obs.market.ids[index])

class Buy(Node):
  priority = 20
  argType  = Fixed

  @staticproperty
  def edges():
    return [MarketItem]

  def enabled(config):
    return config.EXCHANGE_SYSTEM_ENABLED

  def call(realm, entity, item):
    if item is None or item.owner_id.val == 0:
      return

    assert entity.alive, "Dead entity cannot act"
    assert entity.is_player, "Npcs cannot buy an item"
    assert item.quantity.val > 0, "Item quantity cannot be 0" # indicates item leak
    assert item.equipped.val == 0, 'Listed item must not be equipped'

    if not realm.config.EXCHANGE_SYSTEM_ENABLED:
      return

    if entity.gold.val < item.listed_price.val: # not enough money
      return

    if entity.ent_id == item.owner_id.val: # cannot buy own item
      return

    if entity.in_combat: # player cannot buy item during combat
      return

    if not entity.inventory.space:
      # buyer inventory is full - see if it has an ammo stack with the same sig
      if isinstance(item, Stack):
        if not entity.inventory.has_stack(item.signature):
          # no ammo stack with the same signature, so cannot give
          return
      else: # no space, and item is not ammo stack, so cannot give
        return

    # one can try to buy, but the listing might have gone (perhaps bought by other)
    realm.exchange.buy(entity, item)

class Sell(Node):
  priority = 70
  argType  = Fixed

  @staticproperty
  def edges():
    return [InventoryItem, Price]

  def enabled(config):
    return config.EXCHANGE_SYSTEM_ENABLED

  def call(realm, entity, item, price):
    if item is None or item.owner_id.val != entity.ent_id or price is None:
      return

    assert entity.alive, "Dead entity cannot act"
    assert entity.is_player, "Npcs cannot sell an item"
    assert item.quantity.val > 0, "Item quantity cannot be 0" # indicates item leak

    if not realm.config.EXCHANGE_SYSTEM_ENABLED:
      return

    if item not in entity.inventory:
      return

    if entity.in_combat: # player cannot sell item during combat
      return

    # cannot sell the equipped or listed item
    if item.equipped.val or item.listed_price.val:
      return

    if not isinstance(price, int):
      price = price.val

    if not price > 0:
      return

    realm.exchange.sell(entity, item, price, realm.tick)

def init_discrete(values):
  classes = []
  for i in values:
    name = f'Discrete_{i}'
    cls  = type(name, (object,), {'val': i})
    classes.append(cls)

  return classes

class Price(Node):
  argType  = Fixed

  @classmethod
  def init(cls, config):
    # gold should be > 0
    cls.price_range = range(1, config.PRICE_N_OBS+1)
    Price.classes = init_discrete(cls.price_range)

  @classmethod
  def index(cls, price):
    try:
      return cls.price_range.index(price)
    except ValueError:
      # use the max price, which is config.PRICE_N_OBS
      return len(cls.price_range) - 1

  @staticproperty
  def edges():
    return Price.classes

  def deserialize(realm, entity, index, obs: Observation):
    return deserialize_fixed_arg(Price, index)

class Token(Node):
  argType  = Fixed

  @classmethod
  def init(cls, config):
    Token.classes = init_discrete(range(config.COMMUNICATION_NUM_TOKENS))

  @staticproperty
  def edges():
    return Token.classes

  def deserialize(realm, entity, index, obs: Observation):
    return deserialize_fixed_arg(Token, index)

class Comm(Node):
  argType  = Fixed
  priority = 99

  @staticproperty
  def edges():
    return [Token]

  def enabled(config):
    return config.COMMUNICATION_SYSTEM_ENABLED

  def call(realm, entity, token):
    if token is None:
      return

    entity.message.update(token.val)

#TODO: Solve AGI
class BecomeSkynet:
  pass



from __future__ import annotations
from typing import Callable, List, Optional, Tuple, Union, Iterable, Type, TYPE_CHECKING
from types import FunctionType
from abc import ABC, abstractmethod
import inspect
from numbers import Real

from nmmo.core.config import Config
from nmmo.task.group import Group, union
from nmmo.task.game_state import GameState
from nmmo.task.constraint import Constraint, GroupConstraint

if TYPE_CHECKING:
  from nmmo.task.task_api import Task

class InvalidPredicateDefinition(Exception):
  pass

class Predicate(ABC):
  """ A mapping from a game state to bounded [0, 1] float
  """
  def __init__(self,
               subject: Group,
               *args,
               constraints: Optional[List[Tuple[str,Optional[Constraint]]]] = None,
               **kwargs):
    self.name = self._make_name(self.__class__.__name__, args, kwargs)

    self._groups: List[Group] = [x for x in list(args) + list(kwargs.values())
                                 if isinstance(x, Group)]

    self._groups.append(subject)

    self._args = args
    self._kwargs = kwargs
    self._constraints = constraints  # NOTE: not used
    self._config = None
    self._subject = subject

  def __call__(self, gs: GameState) -> float:
    """ Calculates score

    Params:
      gs: GameState

    Returns:
      progress: float bounded between [0, 1], 1 is considered to be true
    """
    # Update views
    for group in self._groups:
      group.update(gs)
    # Calculate score
    cache = gs.cache_result
    if self.name in cache:
      progress = cache[self.name]
    else:
      progress = max(min(self._evaluate(gs)*1.0,1.0),0.0)
      cache[self.name] = progress
    return progress

  def close(self):
    # To prevent memory leak, clear all refs to old game state
    for group in self._groups:
      group.clear_prev_state()

  @abstractmethod
  def _evaluate(self, gs: GameState) -> float:
    """ A mapping from a game state to the desirability/progress of that state.
        __call__() will cap its value to [0, 1]
    """
    raise NotImplementedError

  def _make_name(self, class_name, args, kwargs) -> str:
    name = [class_name] + \
      list(map(arg_to_string, args)) + \
      [f"{arg_to_string(key)}:{arg_to_string(arg)}" for key, arg in kwargs.items()]
    name = "("+'_'.join(name).replace(' ', '')+")"
    return name

  def __str__(self):
    return self.name

  @abstractmethod
  def get_source_code(self) -> str:
    """ Returns the actual source code how the game state/progress evaluation is done.
    """
    raise NotImplementedError

  @abstractmethod
  def get_signature(self) -> List:
    """ Returns the signature of the game state/progress evaluation function.
    """
    raise NotImplementedError

  @property
  def args(self):
    return self._args

  @property
  def kwargs(self):
    return self._kwargs

  @property
  def subject(self):
    return self._subject

  def create_task(self,
                  task_cls: Optional[Type[Task]]=None,
                  assignee: Union[Iterable[int], int]=None,
                  **kwargs) -> Task:
    """ Creates a task from this predicate"""
    if task_cls is None:
      from nmmo.task.task_api import Task
      task_cls = Task

    if assignee is None:
      # the new task is assigned to this predicate's subject
      assignee = self._subject.agents

    return task_cls(eval_fn=self, assignee=assignee, **kwargs)

  def __and__(self, other):
    return AND(self, other)
  def __or__(self, other):
    return OR(self, other)
  def __invert__(self):
    return NOT(self)
  def __add__(self, other):
    return ADD(self, other)
  def __radd__(self, other):
    return ADD(self, other)
  def __sub__(self, other):
    return SUB(self, other)
  def __rsub__(self, other):
    return SUB(self, other)
  def __mul__(self, other):
    return MUL(self, other)
  def __rmul__(self, other):
    return MUL(self, other)

# _make_name helper functions
def arg_to_string(arg):
  if isinstance(arg, (type, FunctionType)): # class or function
    return arg.__name__
  if arg is None:
    return 'Any'
  return str(arg)

################################################

def make_predicate(fn: Callable) -> Type[Predicate]:
  """ Syntactic sugar API for defining predicates from function
  """
  signature = inspect.signature(fn)
  for i, param in enumerate(signature.parameters.values()):
    if i == 0 and param.name != 'gs':
      raise InvalidPredicateDefinition('First parameter must be gs: GameState')
    if i == 1 and (param.name != 'subject'):
      raise InvalidPredicateDefinition("Second parameter must be subject: Group")

  class FunctionPredicate(Predicate):
    def __init__(self, *args, **kwargs) -> None:
      self._signature = signature
      super().__init__(*args, **kwargs)
      self._args = args
      self._kwargs = kwargs
      self.name = self._make_name(fn.__name__, args, kwargs)
    def _evaluate(self, gs: GameState) -> float:
      return fn(gs, *self._args, **self._kwargs)
    def get_source_code(self):
      return inspect.getsource(fn).strip()
    def get_signature(self) -> List:
      return list(self._signature.parameters)

  return FunctionPredicate


################################################
class PredicateOperator(Predicate):
  def __init__(self, n, *predicates: Union[Predicate, Real], subject: Group=None):
    if not n(len(predicates)):
      raise InvalidPredicateDefinition(f"Need {n} arguments")
    predicates = list(predicates)
    self._subject_argument = subject
    if subject is None:
      try:
        subject = union(*[p.subject
                          for p in filter(lambda p: isinstance(p, Predicate), predicates)])
      except AttributeError:
        subject = GroupConstraint()
    super().__init__(subject, *predicates)

    for i, p in enumerate(predicates):
      if isinstance(p, Real):
        predicates[i] = lambda _,v=predicates[i] : v
    self._predicates = predicates

  def check(self, config: Config) -> bool:
    return all((p.check(config) if isinstance(p, Predicate)
                else True for p in self._predicates))

  def sample(self, config: Config, cls: Type[PredicateOperator], **kwargs):
    subject = self._subject_argument if 'subject' not in kwargs else kwargs['subject']
    predicates = [p.sample(config, **kwargs) if isinstance(p, Predicate)
                  else p(None) for p in self._predicates]
    return cls(*predicates, subject=subject)

  def get_source_code(self) -> str:
    # NOTE: get_source_code() of the combined predicates returns the joined str
    #   of each predicate's source code, which may NOT represent what the actual
    #   predicate is doing
    # TODO: try to generate "the source code" that matches
    #   what the actual instantiated predicate returns,
    #   which perhaps should reflect the actual agent ids, etc...
    src_list = []
    for pred in self._predicates:
      if isinstance(pred, Predicate):
        src_list.append(pred.get_source_code())
    return '\n\n'.join(src_list).strip()

  def get_signature(self):
    # TODO: try to generate the correct signature
    return []

  @property
  def args(self):
    # TODO: try to generate the correct args
    return []

  @property
  def kwargs(self):
    # NOTE: This is incorrect implementation. kwargs of the combined predicates returns
    #   all summed kwargs dict, which can OVERWRITE the values of duplicated keys
    # TODO: try to match the eval function and kwargs, which can be correctly used downstream
    # for pred in self._predicates:
    #   if isinstance(pred, Predicate):
    #     kwargs.update(pred.kwargs)
    return {}

class OR(PredicateOperator, Predicate):
  def __init__(self, *predicates: Predicate, subject: Group=None):
    super().__init__(lambda n: n>0, *predicates, subject=subject)
  def _evaluate(self, gs: GameState) -> float:
    # using max as OR for the [0,1] float
    return max(p(gs) for p in self._predicates)
  def sample(self, config: Config, **kwargs):
    return super().sample(config, OR, **kwargs)

class AND(PredicateOperator, Predicate):
  def __init__(self, *predicates: Predicate, subject: Group=None):
    super().__init__(lambda n: n>0, *predicates, subject=subject)
  def _evaluate(self, gs: GameState) -> float:
    # using min as AND for the [0,1] float
    return min(p(gs) for p in self._predicates)
  def sample(self, config: Config, **kwargs):
    return super().sample(config, AND, **kwargs)

class NOT(PredicateOperator, Predicate):
  def __init__(self, predicate: Predicate, subject: Group=None):
    super().__init__(lambda n: n==1, predicate, subject=subject)
  def _evaluate(self, gs: GameState) -> float:
    return 1.0 - self._predicates[0](gs)
  def sample(self, config: Config, **kwargs):
    return super().sample(config, NOT, **kwargs)

class ADD(PredicateOperator, Predicate):
  def __init__(self, *predicate: Union[Predicate, Real], subject: Group=None):
    super().__init__(lambda n: n>0, *predicate, subject=subject)
  def _evaluate(self, gs: GameState) -> float:
    return max(min(sum(p(gs) for p in self._predicates),1.0),0.0)
  def sample(self, config: Config, **kwargs):
    return super().sample(config, ADD, **kwargs)

class SUB(PredicateOperator, Predicate):
  def __init__(self, p: Predicate, q: Union[Predicate, Real], subject: Group=None):
    super().__init__(lambda n: n==2, p,q, subject=subject)
  def _evaluate(self, gs: GameState) -> float:
    return max(min(self._predicates[0](gs)-self._predicates[1](gs),1.0),0.0)
  def sample(self, config: Config, **kwargs):
    return super().sample(config, SUB, **kwargs)

class MUL(PredicateOperator, Predicate):
  def __init__(self, *predicate: Union[Predicate, Real], subject: Group=None):
    super().__init__(lambda n: n>0, *predicate, subject=subject)
  def _evaluate(self, gs: GameState) -> float:
    result = 1.0
    for p in self._predicates:
      result = result * p(gs)
    return max(min(result,1.0),0.0)
  def sample(self, config: Config, **kwargs):
    return super().sample(config, MUL, **kwargs)



class SequentialLoader:
  '''config.PLAYER_LOADER that spreads out agent populations'''
  def __init__(self, config, np_random):
    items = config.PLAYERS

    self.items = items
    self.idx   = -1

    # np_random is the env-level rng
    self.candidate_spawn_pos = spawn_concurrent(config, np_random)

  def __iter__(self):
    return self

  def __next__(self):
    self.idx = (self.idx + 1) % len(self.items)
    return self.items[self.idx]

  # pylint: disable=unused-argument
  def get_spawn_position(self, agent_id):
    # the basic SequentialLoader just provides a random spawn position
    return self.candidate_spawn_pos.pop()

def spawn_continuous(config, np_random):
  '''Generates spawn positions for new agents

  Randomly selects spawn positions around
  the borders of the square game map

  Returns:
      tuple(int, int):

  position:
      The position (row, col) to spawn the given agent
  '''
  #Spawn at edges
  mmax = config.MAP_CENTER + config.MAP_BORDER
  mmin = config.MAP_BORDER

  # np_random is the env-level RNG, a drop-in replacement of numpy.random
  var  = np_random.integers(mmin, mmax)
  fixed = np_random.choice([mmin, mmax])
  r, c = int(var), int(fixed)
  if np_random.random() > 0.5:
    r, c = c, r
  return (r, c)

def get_edge_tiles(config):
  '''Returns a list of all edge tiles'''
  # Accounts for void borders in coord calcs
  left = config.MAP_BORDER
  right = config.MAP_CENTER + config.MAP_BORDER
  lows = config.MAP_CENTER * [left]
  highs = config.MAP_CENTER * [right]
  inc = list(range(config.MAP_BORDER, config.MAP_CENTER+config.MAP_BORDER))

  # All edge tiles in order
  sides = []
  sides.append(list(zip(lows, inc)))
  sides.append(list(zip(inc, highs)))
  sides.append(list(zip(highs, inc[::-1])))
  sides.append(list(zip(inc[::-1], lows)))

  return sides

def spawn_concurrent(config, np_random):
  '''Generates spawn positions for new agents

  Evenly spaces agents around the borders
  of the square game map, assuming the edge tiles are all habitable

  Returns:
      list of tuple(int, int):

  position:
      The position (row, col) to spawn the given agent
  '''
  team_size = config.PLAYER_TEAM_SIZE
  team_n = len(config.PLAYERS)
  teammate_sep = config.PLAYER_SPAWN_TEAMMATE_DISTANCE

  # Number of total border tiles
  total_tiles = 4 * config.MAP_CENTER

  # Number of tiles, including within-team sep, occupied by each team
  tiles_per_team = teammate_sep*(team_size-1) + team_size

  # Number of total tiles dedicated to separating teams
  buffer_tiles = 0
  if team_n > 1:
    buffer_tiles = total_tiles - tiles_per_team*team_n

  # Number of tiles between teams
  team_sep = buffer_tiles // team_n

  sides = []
  for side in get_edge_tiles(config):
    sides += side

  if team_n > 1:
    # Space across and within teams
    spawn_positions = []
    for idx in range(team_sep//2, len(sides), tiles_per_team+team_sep):
      for offset in list(range(0,  tiles_per_team, teammate_sep+1)):
        if len(spawn_positions) >= config.PLAYER_N:
          continue
        pos = sides[idx + offset]
        spawn_positions.append(pos)
  else:
    # team_n = 1: to fit 128 agents in a small map, ignore spacing and spawn randomly
    # np_random is the env-level RNG, a drop-in replacement of numpy.random
    np_random.shuffle(sides)
    spawn_positions = sides[:config.PLAYER_N]

  return spawn_positions



from types import SimpleNamespace
from typing import List
from copy import deepcopy

import numpy as np

from nmmo.datastore.serialized import SerializedState
from nmmo.entity import Entity
from nmmo.systems.item import Item
from nmmo.lib.log import EventCode

# pylint: disable=no-member
EventState = SerializedState.subclass("Event", [
  "recorded", # event_log is write-only, no update or delete, so no need for row id
  "ent_id",
  "tick",

  "event",

  "type",
  "level",
  "number",
  "gold",
  "target_ent",
])

EventAttr = EventState.State.attr_name_to_col

EventState.Query = SimpleNamespace(
  table=lambda ds: ds.table("Event").where_eq(EventAttr["recorded"], 1),
  by_event=lambda ds, event_code: ds.table("Event").where_eq(
    EventAttr["event"], event_code),
  by_tick=lambda ds, tick: ds.table("Event").where_eq(
    EventAttr["tick"], tick),
)

# defining col synoyms for different event types
ATTACK_COL_MAP = {
  'combat_style': EventAttr['type'],
  'damage': EventAttr['number'] }

ITEM_COL_MAP = {
  'item_type': EventAttr['type'],
  'quantity': EventAttr['number'],
  'price': EventAttr['gold'] }

LEVEL_COL_MAP = { 'skill': EventAttr['type'] }

EXPLORE_COL_MAP = { 'distance': EventAttr['number'] }


class EventLogger(EventCode):
  def __init__(self, realm):
    self.realm = realm
    self.config = realm.config
    self.datastore = realm.datastore

    self.valid_events = { val: evt for evt, val in EventCode.__dict__.items()
                           if isinstance(val, int) }
    self._data_by_tick = {}
    self._last_tick = 0
    self._empty_data = np.empty((0, len(EventAttr)))

    # add synonyms to the attributes
    self.attr_to_col = deepcopy(EventAttr)
    self.attr_to_col.update(ATTACK_COL_MAP)
    self.attr_to_col.update(ITEM_COL_MAP)
    self.attr_to_col.update(LEVEL_COL_MAP)
    self.attr_to_col.update(EXPLORE_COL_MAP)

  def reset(self):
    EventState.State.table(self.datastore).reset()

  # define event logging
  def _create_event(self, entity: Entity, event_code: int):
    log = EventState(self.datastore)
    log.recorded.update(1)
    log.ent_id.update(entity.ent_id)
    # the tick increase by 1 after executing all actions
    log.tick.update(self.realm.tick+1)
    log.event.update(event_code)

    return log

  def record(self, event_code: int, entity: Entity, **kwargs):
    if event_code in [EventCode.EAT_FOOD, EventCode.DRINK_WATER,
                      EventCode.GIVE_ITEM, EventCode.DESTROY_ITEM,
                      EventCode.GIVE_GOLD]:
      # Logs for these events are for counting only
      self._create_event(entity, event_code)
      return

    if event_code == EventCode.GO_FARTHEST: # use EXPLORE_COL_MAP
      if ('distance' in kwargs and kwargs['distance'] > 0):
        log = self._create_event(entity, event_code)
        log.number.update(kwargs['distance'])
        return

    if event_code == EventCode.SCORE_HIT:
      # kwargs['combat_style'] should be Skill.CombatSkill
      if ('combat_style' in kwargs and kwargs['combat_style'].SKILL_ID in [1, 2, 3]) & \
         ('damage' in kwargs and kwargs['damage'] >= 0):
        log = self._create_event(entity, event_code)
        log.type.update(kwargs['combat_style'].SKILL_ID)
        log.number.update(kwargs['damage'])
        return

    if event_code == EventCode.PLAYER_KILL:
      if ('target' in kwargs and isinstance(kwargs['target'], Entity)):
        target = kwargs['target']
        log = self._create_event(entity, event_code)
        log.target_ent.update(target.ent_id)

        # CHECK ME: attack_level or "general" level?? need to clarify
        log.level.update(target.attack_level)
        return

    if event_code in [EventCode.CONSUME_ITEM, EventCode.HARVEST_ITEM, EventCode.EQUIP_ITEM,
                      EventCode.LOOT_ITEM]:
      # CHECK ME: item types should be checked. For example,
      #   Only Ration and Potion can be consumed
      #   Only Ration, Potion, Whetstone, Arrow, Runes can be produced
      #   The quantity should be 1 for all of these events
      if ('item' in kwargs and isinstance(kwargs['item'], Item)):
        item = kwargs['item']
        log = self._create_event(entity, event_code)
        log.type.update(item.ITEM_TYPE_ID)
        log.level.update(item.level.val)
        log.number.update(item.quantity.val)
        return

    if event_code in [EventCode.LIST_ITEM, EventCode.BUY_ITEM]:
      if ('item' in kwargs and isinstance(kwargs['item'], Item)) & \
         ('price' in kwargs and kwargs['price'] > 0):
        item = kwargs['item']
        log = self._create_event(entity, event_code)
        log.type.update(item.ITEM_TYPE_ID)
        log.level.update(item.level.val)
        log.number.update(item.quantity.val)
        log.gold.update(kwargs['price'])
        return

    # NOTE: do we want to separate the source of income? from selling vs looting
    if event_code == EventCode.EARN_GOLD:
      if ('amount' in kwargs and kwargs['amount'] > 0):
        log = self._create_event(entity, event_code)
        log.gold.update(kwargs['amount'])
        return

    if event_code == EventCode.LEVEL_UP:
      # kwargs['skill'] should be Skill.Skill
      if ('skill' in kwargs and kwargs['skill'].SKILL_ID in range(1,9)) & \
         ('level' in kwargs and kwargs['level'] >= 0):
        log = self._create_event(entity, event_code)
        log.type.update(kwargs['skill'].SKILL_ID)
        log.level.update(kwargs['level'])
        return

    # If reached here, then something is wrong
    # CHECK ME: The below should be commented out after debugging
    raise ValueError(f"Event code: {event_code}", kwargs)

  def update(self):
    curr_tick = self.realm.tick + 1  # update happens before the tick update
    if curr_tick > self._last_tick:
      self._data_by_tick[curr_tick] = EventState.Query.by_tick(self.datastore, curr_tick)
      self._last_tick = curr_tick

  def get_data(self, event_code=None, agents: List[int]=None, tick: int=None) -> np.ndarray:
    if tick is not None:
      if tick not in self._data_by_tick:
        return self._empty_data
      event_data = self._data_by_tick[tick]
    else:
      event_data = EventState.Query.table(self.datastore)

    if event_data.shape[0] > 0:
      if event_code is None:
        flt_idx = event_data[:, EventAttr["event"]] > 0
      else:
        flt_idx = event_data[:, EventAttr["event"]] == event_code
      if agents:
        flt_idx &= np.in1d(event_data[:, EventAttr["ent_id"]], agents)
      return event_data[flt_idx]

    return self._empty_data



from typing import Dict, List

class TeamHelper():
  def __init__(self, teams: Dict[int, List[int]]):
    self.teams = teams
    self.num_teams = len(teams)
    self.team_size = {}
    self.team_and_position_for_agent = {}
    self.agent_for_team_and_position = {}

    for team_id, team in teams.items():
      self.team_size[team_id] = len(team)
      for position, agent_id in enumerate(team):
        self.team_and_position_for_agent[agent_id] = (team_id, position)
        self.agent_for_team_and_position[team_id, position] = agent_id

  def agent_position(self, agent_id: int) -> int:
    return self.team_and_position_for_agent[agent_id][1]

  def agent_id(self, team_id: int, position: int) -> int:
    return self.agent_for_team_and_position[team_id, position]

  def is_agent_in_team(self, agent_id:int , team_id: int) -> bool:
    return agent_id in self.teams[team_id]

  def get_target_agent(self, team_id: int, target: str):
    team_ids = list(self.teams.keys())
    idx = team_ids.index(team_id)
    if target == "left_team":
      target_id = team_ids[(idx+1) % self.num_teams]
      return self.teams[target_id]
    if target == "left_team_leader":
      target_id = team_ids[(idx+1) % self.num_teams]
      return self.teams[target_id][0]
    if target == "right_team":
      target_id = team_ids[(idx-1) % self.num_teams]
      return self.teams[target_id]
    if target == "right_team_leader":
      target_id = team_ids[(idx-1) % self.num_teams]
      return self.teams[target_id][0]
    if target == "my_team_leader":
      return self.teams[team_id][0]
    if target == "all_foes":
      all_foes = []
      for foe_team_id in team_ids:
        if foe_team_id != team_id:
          all_foes += self.teams[foe_team_id]
      return all_foes
    return None



import functools
from dataclasses import dataclass, field
from typing import Iterable, Dict, List, Union, Type
from types import FunctionType
from copy import deepcopy

import numpy as np

import nmmo
from nmmo.task.task_api import Task, make_same_task
from nmmo.task.predicate_api import Predicate, make_predicate
from nmmo.task.group import Group
from nmmo.task import base_predicates as bp
from nmmo.lib.team_helper import TeamHelper

""" task_spec

    eval_fn can come from the base_predicates.py or could be custom functions like above
    eval_fn_kwargs are the additional args that go into predicate. There are also special keys
      * "target" must be ["left_team", "right_team", "left_team_leader", "right_team_leader"]
          these str will be translated into the actual agent ids

    task_cls specifies the task class to be used. Default is Task.
    task_kwargs are the optional, additional args that go into the task.
 
    reward_to: must be in ["team", "agent"]
      * "team" create a single team task, in which all team members get rewarded
      * "agent" create a task for each agent, in which only the agent gets rewarded
    
    sampling_weight specifies the weight of the task in the curriculum sampling. Default is 1
"""

REWARD_TO = ["agent", "team"]
VALID_TARGET = ["left_team", "left_team_leader",
                "right_team", "right_team_leader",
                "my_team_leader", "all_foes"]

@dataclass
class TaskSpec:
  eval_fn: FunctionType
  eval_fn_kwargs: Dict
  task_cls: Type[Task] = Task
  task_kwargs: Dict = field(default_factory=dict)
  reward_to: str = "agent"
  sampling_weight: float = 1.0
  embedding: np.ndarray = None
  predicate: Predicate = None

  def __post_init__(self):
    if self.predicate is None:
      assert isinstance(self.eval_fn, FunctionType), \
        "eval_fn must be a function"
    else:
      assert self.eval_fn is None, "Cannot specify both eval_fn and predicate"
    assert self.reward_to in REWARD_TO, \
      f"reward_to must be in {REWARD_TO}"
    if "target" in self.eval_fn_kwargs:
      assert self.eval_fn_kwargs["target"] in VALID_TARGET, \
      f"target must be in {VALID_TARGET}"

  @functools.cached_property
  def name(self):
    # pylint: disable=no-member
    kwargs_str = []
    for key, val in self.eval_fn_kwargs.items():
      val_str = str(val)
      if isinstance(val, type):
        val_str = val.__name__
      kwargs_str.append(f"{key}:{val_str}_")
    kwargs_str = "(" + "".join(kwargs_str)[:-1] + ")" # remove the last _
    pred_name = self.eval_fn.__name__ if self.predicate is None else self.predicate.name
    return "_".join([self.task_cls.__name__, pred_name,
                     kwargs_str, "reward_to:" + self.reward_to])

def make_task_from_spec(assign_to: Union[Iterable[int], Dict],
                        task_spec: List[TaskSpec]) -> List[Task]:
  """
  Args:
    assign_to: either a Dict with { team_id: [agent_id]} or a List of agent ids
    task_spec: a list of tuples (reward_to, eval_fn, pred_fn_kwargs, task_kwargs)
    
    each tuple is assigned to the teams
  """
  teams = assign_to
  if not isinstance(teams, Dict): # convert agent id list to the team dict format
    teams = {idx: [agent_id] for idx, agent_id in enumerate(assign_to)}
  team_list = list(teams.keys())
  team_helper = TeamHelper(teams)

  # assign task spec to teams (assign_to)
  tasks = []
  for idx in range(min(len(team_list), len(task_spec))):
    team_id = team_list[idx]

    # map local vars to spec attributes
    reward_to = task_spec[idx].reward_to
    pred_fn = task_spec[idx].eval_fn
    pred_fn_kwargs = deepcopy(task_spec[idx].eval_fn_kwargs)
    task_cls = task_spec[idx].task_cls
    task_kwargs = deepcopy(task_spec[idx].task_kwargs)
    task_kwargs["embedding"] = task_spec[idx].embedding # to pass to task_cls
    task_kwargs["spec_name"] = task_spec[idx].name
    predicate = task_spec[idx].predicate

    # reserve "target" for relative agent mapping
    if "target" in pred_fn_kwargs:
      target = pred_fn_kwargs.pop("target")
      assert target in VALID_TARGET, "Invalid target"
      # translate target to specific agent ids using team_helper
      target = team_helper.get_target_agent(team_id, target)
      pred_fn_kwargs["target"] = target

    # handle some special cases and instantiate the predicate first
    if pred_fn is not None and isinstance(pred_fn, FunctionType):
      # if a function is provided as a predicate
      pred_cls = make_predicate(pred_fn)

    # TODO: should create a test for these
    if (pred_fn in [bp.AllDead]) or \
       (pred_fn in [bp.StayAlive] and "target" in pred_fn_kwargs):
      # use the target as the predicate subject
      pred_fn_kwargs.pop("target") # remove target
      predicate = pred_cls(Group(target), **pred_fn_kwargs)

    # create the task
    if reward_to == "team":
      assignee = team_helper.teams[team_id]
      if predicate is None:
        predicate = pred_cls(Group(assignee), **pred_fn_kwargs)
        tasks.append(predicate.create_task(task_cls=task_cls, **task_kwargs))
      else:
        # this branch is for the cases like AllDead, StayAlive
        tasks.append(predicate.create_task(assignee=assignee, task_cls=task_cls,
                                           **task_kwargs))

    elif reward_to == "agent":
      agent_list = team_helper.teams[team_id]
      if predicate is None:
        tasks += make_same_task(pred_cls, agent_list, pred_kwargs=pred_fn_kwargs,
                                task_cls=task_cls, task_kwargs=task_kwargs)
      else:
        # this branch is for the cases like AllDead, StayAlive
        tasks += [predicate.create_task(assignee=agent_id, task_cls=task_cls, **task_kwargs)
                  for agent_id in agent_list]

  return tasks

# pylint: disable=bare-except,cell-var-from-loop
def check_task_spec(spec_list: List[TaskSpec]) -> List[Dict]:
  teams = {0: [1, 2, 3], 3: [4, 5], 7: [6, 7], 11: [8, 9], 14: [10, 11]}
  config = nmmo.config.Default()
  env = nmmo.Env(config)
  results = []
  for single_spec in spec_list:
    result = {"spec_name": single_spec.name}
    try:
      env.reset(make_task_fn=lambda: make_task_from_spec(teams, [single_spec]))
      for _ in range(3):
        env.step({})
      result["runnable"] = True
    except:
      result["runnable"] = False

    results.append(result)
  return results




from nmmo.systems import item, droptable

class Material:
  capacity = 0
  tool = None
  table = None
  index = None
  respawn = 0

  def __init__(self, config):
    pass

  def __eq__(self, mtl):
    return self.index == mtl.index

  def __equals__(self, mtl):
    return self == mtl

  def harvest(self):
    return self.__class__.table

class Void(Material):
  tex   = 'void'
  index = 0

class Water(Material):
  tex   = 'water'
  index = 1

  table = droptable.Empty()

  def __init__(self, config):
    self.deplete = __class__
    self.respawn  = 1.0

class Grass(Material):
  tex   = 'grass'
  index = 2

class Scrub(Material):
  tex   = 'scrub'
  index = 3

class Foilage(Material):
  tex   = 'foilage'
  index = 4

  deplete = Scrub
  table = droptable.Empty()

  def __init__(self, config):
    if config.RESOURCE_SYSTEM_ENABLED:
      self.capacity = config.RESOURCE_FOILAGE_CAPACITY
      self.respawn  = config.RESOURCE_FOILAGE_RESPAWN

class Stone(Material):
  tex   = 'stone'
  index = 5

class Slag(Material):
  tex   = 'slag'
  index = 6

class Ore(Material):
  tex   = 'ore'
  index = 7

  deplete = Slag
  tool    = item.Pickaxe

  def __init__(self, config):
    cls = self.__class__
    if cls.table is None:
      cls.table = droptable.Standard()
      cls.table.add(item.Whetstone)

      if config.EQUIPMENT_SYSTEM_ENABLED:
        cls.table.add(item.Wand, prob=config.WEAPON_DROP_PROB)

    if config.PROFESSION_SYSTEM_ENABLED:
      self.capacity = config.PROFESSION_ORE_CAPACITY
      self.respawn  = config.PROFESSION_ORE_RESPAWN

  tool    = item.Pickaxe
  deplete = Slag

class Stump(Material):
  tex   = 'stump'
  index = 8

class Tree(Material):
  tex   = 'tree'
  index = 9

  deplete = Stump
  tool    = item.Axe

  def __init__(self, config):
    cls = self.__class__
    if cls.table is None:
      cls.table = droptable.Standard()
      cls.table.add(item.Arrow)
      if config.EQUIPMENT_SYSTEM_ENABLED:
        cls.table.add(item.Spear, prob=config.WEAPON_DROP_PROB)

    if config.PROFESSION_SYSTEM_ENABLED:
      self.capacity = config.PROFESSION_TREE_CAPACITY
      self.respawn  = config.PROFESSION_TREE_RESPAWN

class Fragment(Material):
  tex   = 'fragment'
  index = 10

class Crystal(Material):
  tex   = 'crystal'
  index = 11

  deplete = Fragment
  tool    = item.Chisel

  def __init__(self, config):
    cls = self.__class__
    if cls.table is None:
      cls.table = droptable.Standard()
      cls.table.add(item.Runes)
      if config.EQUIPMENT_SYSTEM_ENABLED:
        cls.table.add(item.Bow, prob=config.WEAPON_DROP_PROB)

    if config.PROFESSION_SYSTEM_ENABLED:
      self.capacity = config.PROFESSION_CRYSTAL_CAPACITY
      self.respawn  = config.PROFESSION_CRYSTAL_RESPAWN

class Weeds(Material):
  tex   = 'weeds'
  index = 12

class Herb(Material):
  tex   = 'herb'
  index = 13

  deplete = Weeds
  tool    = item.Gloves

  table   = droptable.Standard()
  table.add(item.Potion)

  def __init__(self, config):
    if config.PROFESSION_SYSTEM_ENABLED:
      self.capacity = config.PROFESSION_HERB_CAPACITY
      self.respawn  = config.PROFESSION_HERB_RESPAWN

class Ocean(Material):
  tex   = 'ocean'
  index = 14

class Fish(Material):
  tex   = 'fish'
  index = 15

  deplete = Ocean
  tool    = item.Rod

  table   = droptable.Standard()
  table.add(item.Ration)

  def __init__(self, config):
    if config.PROFESSION_SYSTEM_ENABLED:
      self.capacity = config.PROFESSION_FISH_CAPACITY
      self.respawn  = config.PROFESSION_FISH_RESPAWN

# TODO: Fix lint errors
# pylint: disable=all
class Meta(type):
  def __init__(self, name, bases, dict):
    self.indices = {mtl.index for mtl in self.materials}

  def __iter__(self):
    yield from self.materials

  def __contains__(self, mtl):
    if isinstance(mtl, Material):
      mtl = type(mtl)
    if isinstance(mtl, type):
      return mtl in self.materials
    return mtl in self.indices

class All(metaclass=Meta):
  '''List of all materials'''
  materials = {
    Void, Water, Grass, Scrub, Foilage,
    Stone, Slag, Ore, Stump, Tree,
    Fragment, Crystal, Weeds, Herb, Ocean, Fish}

class Impassible(metaclass=Meta):
  '''Materials that agents cannot walk through'''
  materials = {Void, Water, Stone, Ocean, Fish}

class Habitable(metaclass=Meta):
  '''Materials that agents cannot walk on'''
  materials = {Grass, Scrub, Foilage, Ore, Slag, Tree, Stump, Crystal, Fragment, Herb, Weeds}

class Harvestable(metaclass=Meta):
  '''Materials that agents can harvest'''
  materials = {Water, Foilage, Ore, Tree, Crystal, Herb, Fish}



# pylint: disable=invalid-name, unused-argument
import heapq
import numpy as np

from nmmo.core import action
from nmmo.core.observation import Observation
from nmmo.lib import material
from nmmo.systems.ai import utils


def inSight(dr, dc, vision):
  return (-vision <= dr <= vision and
          -vision <= dc <= vision)

def rand(config, ob, actions, np_random):
  direction = np_random.choice(action.Direction.edges)
  actions[action.Move] = {action.Direction: direction}

def towards(direction, np_random):
  if direction == (-1, 0):
    return action.North
  if direction == (1, 0):
    return action.South
  if direction == (0, -1):
    return action.West
  if direction == (0, 1):
    return action.East

  return np_random.choice(action.Direction.edges)

def pathfind(config, ob, actions, rr, cc, np_random):
  direction = aStar(config, ob, actions, rr, cc)
  direction = towards(direction, np_random)
  actions[action.Move] = {action.Direction: direction}

def meander(config, ob, actions, np_random):
  cands = []
  if ob.tile(-1, 0).material_id in material.Habitable.indices:
    cands.append((-1, 0))
  if ob.tile(1, 0).material_id in material.Habitable.indices:
    cands.append((1, 0))
  if ob.tile(0, -1).material_id in material.Habitable.indices:
    cands.append((0, -1))
  if ob.tile(0, 1).material_id in material.Habitable.indices:
    cands.append((0, 1))

  if len(cands) > 0:
    direction = np_random.choices(cands)[0]
    direction = towards(direction, np_random)
    actions[action.Move] = {action.Direction: direction}

def explore(config, ob, actions, r, c, np_random):
  vision = config.PLAYER_VISION_RADIUS
  sz     = config.MAP_SIZE
  centR, centC = sz//2, sz//2
  vR, vC = centR-r, centC-c
  mmag = max(1, abs(vR), abs(vC))
  rr   = int(np.round(vision*vR/mmag))
  cc   = int(np.round(vision*vC/mmag))
  pathfind(config, ob, actions, rr, cc, np_random)

def evade(config, ob: Observation, actions, attacker, np_random):
  agent = ob.agent()
  rr, cc = (2*agent.row - attacker.row, 2*agent.col - attacker.col)
  pathfind(config, ob, actions, rr, cc, np_random)

def forageDijkstra(config, ob: Observation, actions,
                   food_max, water_max, np_random, cutoff=100):
  vision = config.PLAYER_VISION_RADIUS

  agent  = ob.agent()
  food = agent.food
  water = agent.water

  best      = -1000
  start     = (0, 0)
  goal      = (0, 0)

  reward    = {start: (food, water)}
  backtrace = {start: None}

  queue = [start]

  while queue:
    cutoff -= 1
    if cutoff <= 0:
      break

    cur = queue.pop(0)
    for nxt in utils.adjacentPos(cur):
      if nxt in backtrace:
        continue

      if not inSight(*nxt, vision):
        continue

      tile     = ob.tile(*nxt)
      matl     = tile.material_id

      if not matl in material.Habitable.indices:
        continue

      food, water = reward[cur]
      water = max(0, water - 1)
      food  = max(0, food - 1)
      if matl == material.Foilage.index:
        food = min(food+food_max//2, food_max)

      for pos in utils.adjacentPos(nxt):
        if not inSight(*pos, vision):
          continue

        tile = ob.tile(*pos)
        matl = tile.material_id
        if matl == material.Water.index:
          water = min(water+water_max//2, water_max)
          break

      reward[nxt] = (food, water)

      total = min(food, water)
      if total > best \
          or (total == best and max(food, water) > max(reward[goal])):
        best = total
        goal = nxt

      queue.append(nxt)
      backtrace[nxt] = cur

  while goal in backtrace and backtrace[goal] != start:
    goal = backtrace[goal]
  direction = towards(goal, np_random)
  actions[action.Move] = {action.Direction: direction}

def findResource(config, ob: Observation, resource):
  vision = config.PLAYER_VISION_RADIUS
  resource_index = resource.index
  for r in range(-vision, vision+1):
    for c in range(-vision, vision+1):
      tile = ob.tile(r, c)
      material_id = tile.material_id
    if material_id == resource_index:
      return (r, c)
  return False

def gatherAStar(config, ob, actions, resource, np_random, cutoff=100):
  resource_pos = findResource(config, ob, resource)
  if not resource_pos:
    return False

  rr, cc = resource_pos
  next_pos = aStar(config, ob, actions, rr, cc, cutoff=cutoff)
  if not next_pos or next_pos == (0, 0):
    return False

  direction = towards(next_pos, np_random)
  actions[action.Move] = {action.Direction: direction}
  return True

def gatherBFS(config, ob: Observation, actions, resource, np_random, cutoff=100):
  vision = config.PLAYER_VISION_RADIUS

  start  = (0, 0)
  backtrace = {start: None}
  queue = [start]
  found = False

  while queue:
    cutoff -= 1
    if cutoff <= 0:
      return False

    cur = queue.pop(0)
    for nxt in utils.adjacentPos(cur):
      if found:
        break

      if nxt in backtrace:
        continue

      if not inSight(*nxt, vision):
        continue

      tile     = ob.tile(*nxt)
      matl     = tile.material_id

      if material.Fish in resource and material.Fish.index == matl:
        found = nxt
        backtrace[nxt] = cur
        break

      if not tile.material_id in material.Habitable.indices:
        continue

      if matl in (e.index for e in resource):
        found = nxt
        backtrace[nxt] = cur
        break

      for pos in utils.adjacentPos(nxt):
        if not inSight(*pos, vision):
          continue

        tile = ob.tile(*pos)
        matl = tile.material_id

        if matl == material.Fish.index:
          backtrace[nxt] = cur
          break

        queue.append(nxt)
        backtrace[nxt] = cur

  #Ran out of tiles
  if not found:
    return False

  while found in backtrace and backtrace[found] != start:
    found = backtrace[found]

  direction = towards(found, np_random)
  actions[action.Move] = {action.Direction: direction}

  return True


def aStar(config, ob: Observation, actions, rr, cc, cutoff=100):
  vision = config.PLAYER_VISION_RADIUS

  start = (0, 0)
  goal  = (rr, cc)
  if start == goal:
    return (0, 0)

  pq = [(0, start)]

  backtrace = {}
  cost = {start: 0}

  closestPos = start
  closestHeuristic = utils.l1(start, goal)
  closestCost = closestHeuristic

  while pq:
    # Use approximate solution if budget exhausted
    cutoff -= 1
    if cutoff <= 0:
      if goal not in backtrace:
        goal = closestPos
      break

    priority, cur = heapq.heappop(pq)

    if cur == goal:
      break

    for nxt in utils.adjacentPos(cur):
      if not inSight(*nxt, vision):
        continue

      tile     = ob.tile(*nxt)
      matl     = tile.material_id

      if not matl in material.Habitable.indices:
        continue

      #Omitted water from the original implementation. Seems key
      if matl in material.Impassible.indices:
        continue

      newCost = cost[cur] + 1
      if nxt not in cost or newCost < cost[nxt]:
        cost[nxt] = newCost
        heuristic = utils.lInfty(goal, nxt)
        priority = newCost + heuristic

        # Compute approximate solution
        if heuristic < closestHeuristic \
            or (heuristic == closestHeuristic and priority < closestCost):
          closestPos = nxt
          closestHeuristic = heuristic
          closestCost = priority

        heapq.heappush(pq, (priority, nxt))
        backtrace[nxt] = cur

  goal = closestPos
  while goal in backtrace and backtrace[goal] != start:
    goal = backtrace[goal]

  return goal




from nmmo.systems.ai import behavior, utils

def passive(realm, entity):
  behavior.update(entity)
  actions = {}

  behavior.meander(realm, actions, entity)

  return actions

def neutral(realm, entity):
  behavior.update(entity)
  actions = {}

  if not entity.attacker:
    behavior.meander(realm, actions, entity)
  else:
    entity.target = entity.attacker
    behavior.hunt(realm, actions, entity)

  return actions

def hostile(realm, entity):
  behavior.update(entity)
  actions = {}

  # This is probably slow
  if not entity.target:
    entity.target = utils.closestTarget(entity, realm.map.tiles,
                                        rng=entity.vision)

  if not entity.target:
    behavior.meander(realm, actions, entity)
  else:
    behavior.hunt(realm, actions, entity)

  return actions





import nmmo
from scripted import baselines

def test_pettingzoo_api():
  config = nmmo.config.Default()
  config.PLAYERS = [baselines.Random]
  # ensv = nmmo.Env(config)
  # TODO: disabled due to Env not implementing the correct PettinZoo step() API
  # parallel_api_test(env, num_cycles=1000)


if __name__ == '__main__':
  test_pettingzoo_api()



from ordered_set import OrderedSet

class IdAllocator:
  def __init__(self, max_id):
    # Key 0 is reserved as padding
    self.max_id = 1
    self.free = OrderedSet()
    self.expand(max_id)

  def full(self):
    return len(self.free) == 0

  def remove(self, row_id):
    self.free.add(row_id)

  def allocate(self):
    return self.free.pop(0)

  def expand(self, max_id):
    self.free.update(range(self.max_id, max_id))
    self.max_id = max_id



import numpy as np

from nmmo.lib.colors import Neon
from nmmo.systems import combat

from .render_utils import normalize

# pylint: disable=unused-argument
class OverlayRegistry:
  def __init__(self, realm, renderer):
    '''Manager class for overlays

    Args:
        config: A Config object
        realm: An environment
    '''
    self.initialized = False

    self.realm  = realm
    self.config = realm.config
    self.renderer = renderer

    self.overlays = {
       #'counts':     Counts, # TODO: change population to team
       'skills':     Skills}

  def init(self, *args):
    self.initialized = True
    for cmd, overlay in self.overlays.items():
      self.overlays[cmd] = overlay(self.config, self.realm, self.renderer, *args)
    return self

  def step(self, cmd):
    '''Per-tick overlay updates

    Args:
        cmd: User command returned by the client
    '''
    if not self.initialized:
      self.init()

    for overlay in self.overlays.values():
      overlay.update()

    if cmd in self.overlays:
      self.overlays[cmd].register()


class Overlay:
  '''Define a overlay for visualization in the client

  Overlays are color images of the same size as the game map.
  They are rendered over the environment with transparency and
  can be used to gain insight about agent behaviors.'''
  def __init__(self, config, realm, renderer, *args):
    '''
    Args:
        config: A Config object
        realm: An environment
    '''
    self.config     = config
    self.realm      = realm
    self.renderer   = renderer

    self.size       = config.MAP_SIZE
    self.values     = np.zeros((self.size, self.size))

  def update(self):
    '''Compute per-tick updates to this overlay. Override per overlay.

    Args:
        obs: Observation returned by the environment
    '''

  def register(self):
    '''Compute the overlay and register it within realm. Override per overlay.'''


class Skills(Overlay):
  def __init__(self, config, realm, renderer, *args):
    '''Indicates whether agents specialize in foraging or combat'''
    super().__init__(config, realm, renderer)
    self.num_skill = 2

    self.values  = np.zeros((self.size, self.size, self.num_skill))

  def update(self):
    '''Computes a count-based exploration map by painting
    tiles as agents walk over them'''
    for agent in self.realm.players.values():
      r, c = agent.pos

      skill_lvl  = (agent.skills.food.level.val + agent.skills.water.level.val)/2.0
      combat_lvl = combat.level(agent.skills)

      if skill_lvl == 10 and combat_lvl == 3:
        continue

      self.values[r, c, 0] = skill_lvl
      self.values[r, c, 1] = combat_lvl

  def register(self):
    values = np.zeros((self.size, self.size, self.num_skill))
    for idx in range(self.num_skill):
      ary  = self.values[:, :, idx]
      vals = ary[ary != 0]
      mean = np.mean(vals)
      std  = np.std(vals)
      if std == 0:
        std = 1

      values[:, :, idx] = (ary - mean) / std
      values[ary == 0] = 0

    colors    = np.array([Neon.BLUE.rgb, Neon.BLOOD.rgb])
    colorized = np.zeros((self.size, self.size, 3))
    amax      = np.argmax(values, -1)

    for idx in range(self.num_skill):
      colorized[amax == idx] = colors[idx] / 255
      colorized[values[:, :, idx] == 0] = 0

    self.renderer.register(colorized)


# CHECK ME: this was based on population, so disabling it for now
#   We may want this back for the team-level analysis
class Counts(Overlay):
  def __init__(self, config, realm, renderer, *args):
    super().__init__(config, realm, renderer)
    self.values = np.zeros((self.size, self.size, config.PLAYER_POLICIES))

  def update(self):
    '''Computes a count-based exploration map by painting
    tiles as agents walk over them'''
    for ent_id, agent in self.realm.players.items():
      r, c = agent.pos
      self.values[r, c][ent_id] += 1

  def register(self):
    colors    = self.realm.players.palette.colors
    colors    = np.array([colors[pop].rgb
                          for pop in range(self.config.PLAYER_POLICIES)])

    colorized = self.values[:, :, :, None] * colors / 255
    colorized = np.sum(colorized, -2)
    count_sum  = np.sum(self.values[:, :], -1)
    data      = normalize(count_sum)[..., None]

    count_sum[count_sum==0] = 1
    colorized = colorized * data / count_sum[..., None]

    self.renderer.register(colorized)



from __future__ import annotations
import numpy as np

from nmmo.render import websocket
from nmmo.render.overlay import OverlayRegistry
from nmmo.render.render_utils import patch_packet


# Render is external to the game
class WebsocketRenderer:
  def __init__(self, realm=None) -> None:
    self._client = websocket.Application(realm)
    self.overlay_pos = [256, 256]

    self._realm = realm

    self.overlay = None
    self.registry = OverlayRegistry(realm, renderer=self) if realm else None

    self.packet = None

  def set_realm(self, realm) -> None:
    self._realm = realm
    self.registry = OverlayRegistry(realm, renderer=self) if realm else None

  def render_packet(self, packet) -> None:
    packet = {
      'pos': self.overlay_pos,
      'wilderness': 0, # obsolete, but maintained for compatibility
      **packet }

    self.overlay_pos, _ = self._client.update(packet)

  def render_realm(self) -> None:
    assert self._realm is not None, 'This function requires a realm'
    assert self._realm.tick is not None, 'render before reset'

    packet = {
      'config': self._realm.config,
      'pos': self.overlay_pos,
      'wilderness': 0,
      **self._realm.packet()
    }

    # TODO: a hack to make the client work
    packet = patch_packet(packet, self._realm)

    if self.overlay is not None:
      packet['overlay'] = self.overlay
      self.overlay = None

    # save the packet for investigation
    self.packet = packet

    # pass the packet to renderer
    pos, cmd = self._client.update(self.packet)

    self.overlay_pos = pos
    self.registry.step(cmd)

  def register(self, overlay: np.ndarray) -> None:
    '''Register an overlay to be sent to the client

    The intended use of this function is: User types overlay ->
    client sends cmd to server -> server computes overlay update ->
    register(overlay) -> overlay is sent to client -> overlay rendered

    Args:
        overlay: A map-sized (self.size) array of floating point values
        overlay must be a numpy array of dimension (*(env.size), 3)
    '''
    self.overlay = overlay.tolist()



class Fixed():
  def __init__(self, item):
    self.item = item

  def roll(self, realm, level):
    return [self.item(realm, level)]

class Drop:
  def __init__(self, item, prob):
    self.item = item
    self.prob = prob

  def roll(self, realm, level):
    # TODO: do not access realm._np_random directly
    #   related to skill.py, all harvest skills
    # pylint: disable=protected-access
    if realm._np_random.random() < self.prob:
      return self.item(realm, level)

    return None

class Standard:
  def __init__(self):
    self.drops = []

  def add(self, item, prob=1.0):
    self.drops += [Drop(item, prob)]

  def roll(self, realm, level):
    ret = []
    for e in self.drops:
      drop = e.roll(realm, level)
      if drop is not None:
        ret += [drop]
    return ret

class Empty(Standard):
  def roll(self, realm, level):
    return []

class Ammunition(Standard):
  def __init__(self, item):
    super().__init__()
    self.item = item

  def roll(self, realm, level):
    return [self.item(realm, level)]

class Consumable(Standard):
  def __init__(self, item):
    super().__init__()
    self.item = item

  def roll(self, realm, level):
    return [self.item(realm, level)]



import unittest

from nmmo.datastore.id_allocator import IdAllocator

class TestIdAllocator(unittest.TestCase):
  def test_id_allocator(self):
    id_allocator = IdAllocator(10)

    for i in range(1, 10):
      row_id = id_allocator.allocate()
      self.assertEqual(i, row_id)
    self.assertTrue(id_allocator.full())

    id_allocator.remove(5)
    id_allocator.remove(6)
    id_allocator.remove(1)
    self.assertFalse(id_allocator.full())

    self.assertSetEqual(
      set(id_allocator.allocate() for i in range(3)),
      set([5, 6, 1])
    )
    self.assertTrue(id_allocator.full())

    id_allocator.expand(11)
    self.assertFalse(id_allocator.full())

    self.assertEqual(id_allocator.allocate(), 10)

    with self.assertRaises(KeyError):
      id_allocator.allocate()

  def test_id_reuse(self):
    id_allocator = IdAllocator(10)

    for i in range(1, 10):
      row_id = id_allocator.allocate()
      self.assertEqual(i, row_id)
    self.assertTrue(id_allocator.full())

    id_allocator.remove(5)
    id_allocator.remove(6)
    id_allocator.remove(1)
    self.assertFalse(id_allocator.full())

    self.assertSetEqual(
      set(id_allocator.allocate() for i in range(3)),
      set([5, 6, 1])
    )
    self.assertTrue(id_allocator.full())

    id_allocator.expand(11)
    self.assertFalse(id_allocator.full())

    self.assertEqual(id_allocator.allocate(), 10)

    with self.assertRaises(KeyError):
      id_allocator.allocate()

    id_allocator.remove(10)
    self.assertEqual(id_allocator.allocate(), 10)

if __name__ == '__main__':
  unittest.main()



from __future__ import annotations

import random
from numbers import Number
from typing import Union, Callable, Dict
from abc import ABC, abstractmethod

from nmmo.systems import skill, item
from nmmo.lib import material
from nmmo.lib.log import EventCode
from nmmo.core.config import Config

class InvalidConstraint(Exception):
  pass

class Constraint(ABC):
  """ To check the validity of predicates
  and assist generating new predicates. Similar to gym spaces.
  """
  def __init__(self, systems=None):
    if systems is None:
      systems = []
    self._systems = systems

  # pylint: disable=unused-argument
  def check(self, config: Config, value):
    """ Checks value is in bounds given config
    """
    for system in self._systems:
      try:
        if not getattr(config,system):
          return False
      except AttributeError:
        return False
    return True

  @abstractmethod
  def sample(self, config: Config):
    """ Generator to sample valid values given config
    """
    raise NotImplementedError

  def __str__(self):
    return self.__class__.__name__

# This is a dummy function for GroupConstraint
# NOTE: config does not have team info
def sample_one_big_team(config):
  from nmmo.task.group import Group
  team = list(range(1, config.PLAYER_N+1))
  return [Group(team, 'All')]

class GroupConstraint(Constraint):
  """ Ensures that all agents of a group exist in a config
  """
  def __init__(self,
               sample_fn = sample_one_big_team,
               systems = None):
    """
    Params
      sample_fn: given a Config, return groups to select from
      systems: systems required to operate
    """
    super().__init__(systems)
    self._sample_fn = sample_fn

  def check(self, config, value):
    if not super().check(config,value):
      return False
    for agent in value.agents:
      if agent > config.PLAYER_N:
        return False
    return True

  def sample(self, config):
    return random.choice(self._sample_fn(config))

  def sample_from_teams(self, teams: Dict[int, Dict]):
    from nmmo.task.group import Group
    team_id = random.choice(list(teams.keys()))
    return Group(teams[team_id], str(team_id))

class AgentListConstraint(Constraint):
  """ Ensures that all agents of the list exist in a config
  """
  def check(self, config, value):
    for agent in value:
      if agent > config.PLAYER_N:
        return False
    return True

  def sample(self, config):
    return None

class ScalarConstraint(Constraint):
  def __init__(self,
               low:  Union[Callable, Number] = 0,
               high: Union[Callable, Number] = 1024,
               dtype = int,
               systems = None):
    super().__init__(systems)
    self._low = low
    self._high = high
    if isinstance(low, Number):
      self._low  = lambda _ : low
    if isinstance(high, Number):
      self._high = lambda _ : high
    self._dtype = dtype

  def check(self, config, value):
    if not super().check(config,value):
      return False
    if self._low(config) <= value < self._high(config):
      return True
    return False

  def sample(self, config):
    l, h = self._low(config), self._high(config)
    return self._dtype(random.random()*(h-l)+l)

class DiscreteConstraint(Constraint):
  def __init__(self, space, systems=None):
    super().__init__(systems)
    self._space = set(space)

  def check(self, config: Config, value):
    if not super().check(config,value):
      return False
    return value in self._space

  def sample(self, config: Config):
    # NOTE: this does NOT need to be deterministic
    return random.choice(self._space)

# Group Constraints
TEAM_GROUPS = GroupConstraint()
INDIVIDUAL_GROUPS=GroupConstraint()
AGENT_LIST_CONSTRAINT = AgentListConstraint()

# Tile Constraints
MATERIAL_CONSTRAINT = DiscreteConstraint(space=list(material.All.materials),
                                         systems=['TERRAIN_SYSTEM_ENABLED',
                                                  'RESOURCE_SYSTEM_ENABLED'])
HABITABLE_CONSTRAINT = DiscreteConstraint(space=list(material.Habitable.materials),
                                         systems=['TERRAIN_SYSTEM_ENABLED'])

# Event Constraints
event_names = [k for k, v in EventCode.__dict__.items() if isinstance(v,int)]
EVENTCODE_CONSTRAINT = DiscreteConstraint(space=event_names)

# Skill Constraints
combat_skills = [skill.Melee, skill.Mage, skill.Range]
harvest_skills = [skill.Fishing, skill.Herbalism, skill.Prospecting, skill.Alchemy, skill.Carving]
SKILL_CONSTRAINT = DiscreteConstraint(space=combat_skills+harvest_skills,
                                      systems=['PROFESSION_SYSTEM_ENABLED'])
COMBAT_SKILL_CONSTRAINT = DiscreteConstraint(space=combat_skills,
                                      systems=['PROFESSION_SYSTEM_ENABLED'])

# Item Constraints
armour = [item.Hat, item.Top, item.Bottom]
weapons = [item.Spear, item.Bow, item.Wand]
tools = [item.Axe, item.Gloves, item.Rod, item.Pickaxe, item.Chisel]
ammunition = [item.Runes, item.Arrow, item.Whetstone]
consumables = [item.Potion, item.Ration]
ITEM_CONSTRAINT = DiscreteConstraint(space=armour+weapons+tools+ammunition+consumables,
                                     systems=['ITEM_SYSTEM_ENABLED'])
EQUIPABLE_CONSTRAINT = DiscreteConstraint(space=armour+weapons+tools+ammunition,
                                     systems=['ITEM_SYSTEM_ENABLED'])
CONSUMABLE_CONSTRAINT = DiscreteConstraint(space=consumables,
                                           systems=['ITEM_SYSTEM_ENABLED'])
HARVEST_CONSTRAINT = DiscreteConstraint(space=weapons+ammunition+consumables,
                                        systems=['ITEM_SYSTEM_ENABLED'])

# Config Constraints
COORDINATE_CONSTRAINT = ScalarConstraint(high = lambda c: c.MAP_CENTER)
PROGRESSION_CONSTRAINT = ScalarConstraint(high = lambda c: c.PROGRESSION_LEVEL_MAX+1)
INVENTORY_CONSTRAINT = ScalarConstraint(high=lambda c: c.ITEM_INVENTORY_CAPACITY+1)
AGENT_NUMBER_CONSTRAINT = ScalarConstraint(low = 1, high = lambda c: c.PLAYER_N+1)

# Arbitrary Constraints
EVENT_NUMBER_CONSTRAINT = ScalarConstraint(low = 1, high = 110)
GOLD_CONSTRAINT = ScalarConstraint(low = 1, high = 1000)
AGENT_TYPE_CONSTRAINT = DiscreteConstraint(space=['npc','player'])



import unittest

import nmmo
from nmmo.datastore.numpy_datastore import NumpyDatastore
from nmmo.lib.event_log import EventState, EventLogger
from nmmo.lib.log import EventCode
from nmmo.entity.entity import Entity
from nmmo.systems.item import ItemState
from nmmo.systems.item import Whetstone, Ration, Hat
from nmmo.systems import skill as Skill


class MockRealm:
  def __init__(self):
    self.config = nmmo.config.Default()
    self.datastore = NumpyDatastore()
    self.items = {}
    self.datastore.register_object_type("Event", EventState.State.num_attributes)
    self.datastore.register_object_type("Item", ItemState.State.num_attributes)
    self.tick = 0


class MockEntity(Entity):
  # pylint: disable=super-init-not-called
  def __init__(self, ent_id, **kwargs):
    self.id = ent_id
    self.level = kwargs.pop('attack_level', 0)

  @property
  def ent_id(self):
    return self.id

  @property
  def attack_level(self):
    return self.level


class TestEventLog(unittest.TestCase):

  def test_event_logging(self):
    mock_realm = MockRealm()
    event_log = EventLogger(mock_realm)

    mock_realm.tick = 0 # tick increase to 1 after all actions are processed
    event_log.record(EventCode.EAT_FOOD, MockEntity(1))
    event_log.record(EventCode.DRINK_WATER, MockEntity(2))
    event_log.record(EventCode.SCORE_HIT, MockEntity(2),
                     combat_style=Skill.Melee, damage=50)
    event_log.record(EventCode.PLAYER_KILL, MockEntity(3),
                     target=MockEntity(5, attack_level=5))
    event_log.update()

    mock_realm.tick = 1
    event_log.record(EventCode.CONSUME_ITEM, MockEntity(4),
                     item=Ration(mock_realm, 8))
    event_log.record(EventCode.GIVE_ITEM, MockEntity(4))
    event_log.record(EventCode.DESTROY_ITEM, MockEntity(5))
    event_log.record(EventCode.HARVEST_ITEM, MockEntity(6),
                     item=Whetstone(mock_realm, 3))
    event_log.update()

    mock_realm.tick = 2
    event_log.record(EventCode.GIVE_GOLD, MockEntity(7))
    event_log.record(EventCode.LIST_ITEM, MockEntity(8),
                     item=Ration(mock_realm, 5), price=11)
    event_log.record(EventCode.EARN_GOLD, MockEntity(9), amount=15)
    event_log.record(EventCode.BUY_ITEM, MockEntity(10),
                     item=Whetstone(mock_realm, 7), price=21)
    #event_log.record(EventCode.SPEND_GOLD, env.realm.players[11], amount=25)
    event_log.update()

    mock_realm.tick = 3
    event_log.record(EventCode.LEVEL_UP, MockEntity(12),
                     skill=Skill.Fishing, level=3)
    event_log.update()

    mock_realm.tick = 4
    event_log.record(EventCode.GO_FARTHEST, MockEntity(12), distance=6)
    event_log.record(EventCode.EQUIP_ITEM, MockEntity(12),
                     item=Hat(mock_realm, 4))
    event_log.update()

    log_data = [list(row) for row in event_log.get_data()]
    self.assertListEqual(log_data, [
      [1,  1, 1, EventCode.EAT_FOOD, 0, 0, 0, 0, 0],
      [1,  2, 1, EventCode.DRINK_WATER, 0, 0, 0, 0, 0],
      [1,  2, 1, EventCode.SCORE_HIT, 1, 0, 50, 0, 0],
      [1,  3, 1, EventCode.PLAYER_KILL, 0, 5, 0, 0, 5],
      [1,  4, 2, EventCode.CONSUME_ITEM, 16, 8, 1, 0, 0],
      [1,  4, 2, EventCode.GIVE_ITEM, 0, 0, 0, 0, 0],
      [1,  5, 2, EventCode.DESTROY_ITEM, 0, 0, 0, 0, 0],
      [1,  6, 2, EventCode.HARVEST_ITEM, 13, 3, 1, 0, 0],
      [1,  7, 3, EventCode.GIVE_GOLD, 0, 0, 0, 0, 0],
      [1,  8, 3, EventCode.LIST_ITEM, 16, 5, 1, 11, 0],
      [1,  9, 3, EventCode.EARN_GOLD, 0, 0, 0, 15, 0],
      [1, 10, 3, EventCode.BUY_ITEM, 13, 7, 1, 21, 0],
      [1, 12, 4, EventCode.LEVEL_UP, 4, 3, 0, 0, 0],
      [1, 12, 5, EventCode.GO_FARTHEST, 0, 0, 6, 0, 0],
      [1, 12, 5, EventCode.EQUIP_ITEM, 2, 4, 1, 0, 0]])

    log_by_tick = [list(row) for row in event_log.get_data(tick = 4)]
    self.assertListEqual(log_by_tick, [
      [1, 12, 4, EventCode.LEVEL_UP, 4, 3, 0, 0, 0]])

    log_by_event = [list(row) for row in event_log.get_data(event_code = EventCode.CONSUME_ITEM)]
    self.assertListEqual(log_by_event, [
      [1,  4, 2, EventCode.CONSUME_ITEM, 16, 8, 1, 0, 0]])

    log_by_tick_agent = [list(row) for row in \
                         event_log.get_data(tick = 5,
                                            agents = [12],
                                            event_code = EventCode.EQUIP_ITEM)]
    self.assertListEqual(log_by_tick_agent, [
      [1, 12, 5, EventCode.EQUIP_ITEM, 2, 4, 1, 0, 0]])

    empty_log = event_log.get_data(tick = 10)
    self.assertTrue(empty_log.shape[0] == 0)

if __name__ == '__main__':
  unittest.main()

  """
  TEST_HORIZON = 50
  RANDOM_SEED = 338

  from tests.testhelpers import ScriptedAgentTestConfig, ScriptedAgentTestEnv

  config = ScriptedAgentTestConfig()
  env = ScriptedAgentTestEnv(config)

  env.reset(seed=RANDOM_SEED)

  from tqdm import tqdm
  for tick in tqdm(range(TEST_HORIZON)):
    env.step({})

    # events to check
    log = env.realm.event_log.get_data()    
    idx = (log[:,2] == tick+1) & (log[:,3] == EventCode.EQUIP_ITEM)
    if sum(idx):
      print(log[idx])
      print()

  print('done')
  """



#Various utilities for managing combat, including hit/damage

import numpy as np

from nmmo.systems import skill as Skill
from nmmo.lib.log import EventCode

def level(skills):
  return max(e.level.val for e in skills.skills)

def damage_multiplier(config, skill, targ):
  skills = [targ.skills.melee, targ.skills.range, targ.skills.mage]
  exp    = [s.exp for s in skills]

  if max(exp) == min(exp):
    return 1.0

  idx    = np.argmax([exp])
  targ   = skills[idx]

  if isinstance(skill, targ.weakness):
    return config.COMBAT_WEAKNESS_MULTIPLIER

  return 1.0

# pylint: disable=unnecessary-lambda-assignment
def attack(realm, player, target, skill_fn):
  config       = player.config
  skill        = skill_fn(player)
  skill_type   = type(skill)
  skill_name   = skill_type.__name__

  # Per-style offense/defense
  level_damage = 0
  if skill_type == Skill.Melee:
    base_damage  = config.COMBAT_MELEE_DAMAGE

    if config.PROGRESSION_SYSTEM_ENABLED:
      base_damage  = config.PROGRESSION_MELEE_BASE_DAMAGE
      level_damage = config.PROGRESSION_MELEE_LEVEL_DAMAGE

    offense_fn   = lambda e: e.melee_attack
    defense_fn   = lambda e: e.melee_defense

  elif skill_type == Skill.Range:
    base_damage  = config.COMBAT_RANGE_DAMAGE

    if config.PROGRESSION_SYSTEM_ENABLED:
      base_damage  = config.PROGRESSION_RANGE_BASE_DAMAGE
      level_damage = config.PROGRESSION_RANGE_LEVEL_DAMAGE

    offense_fn   = lambda e: e.range_attack
    defense_fn   = lambda e: e.range_defense

  elif skill_type == Skill.Mage:
    base_damage  = config.COMBAT_MAGE_DAMAGE

    if config.PROGRESSION_SYSTEM_ENABLED:
      base_damage  = config.PROGRESSION_MAGE_BASE_DAMAGE
      level_damage = config.PROGRESSION_MAGE_LEVEL_DAMAGE

    offense_fn   = lambda e: e.mage_attack
    defense_fn   = lambda e: e.mage_defense

  elif __debug__:
    assert False, 'Attack skill must be Melee, Range, or Mage'

  # Compute modifiers
  multiplier        = damage_multiplier(config, skill, target)
  skill_offense     = base_damage + level_damage * skill.level.val

  if config.PROGRESSION_SYSTEM_ENABLED:
    skill_defense     = config.PROGRESSION_BASE_DEFENSE  + \
      config.PROGRESSION_LEVEL_DEFENSE*level(target.skills)
  else:
    skill_defense     = 0

  if config.EQUIPMENT_SYSTEM_ENABLED:
    equipment_offense = player.equipment.total(offense_fn)
    equipment_defense = target.equipment.total(defense_fn)

    # after tallying ammo damage, consume ammo (i.e., fire) when the skill type matches
    ammunition = player.equipment.ammunition.item
    if ammunition is not None and getattr(ammunition, skill_name.lower() + '_attack').val > 0:
      ammunition.fire(player)

  else:
    equipment_offense = 0
    equipment_defense = 0

  # Total damage calculation
  offense = skill_offense + equipment_offense
  defense = skill_defense + equipment_defense
  damage  = config.COMBAT_DAMAGE_FORMULA(offense, defense, multiplier)
  #damage  = multiplier * (offense - defense)
  damage  = max(int(damage), 0)

  if player.is_player:
    equipment_level_offense = 0
    equipment_level_defense = 0
    if config.EQUIPMENT_SYSTEM_ENABLED:
      equipment_level_offense = player.equipment.total(lambda e: e.level)
      equipment_level_defense = target.equipment.total(lambda e: e.level)

    realm.event_log.record(EventCode.SCORE_HIT, player,
                           combat_style=skill_type, damage=damage)

    realm.log_milestone(f'Damage_{skill_name}', damage,
                        f'COMBAT: Inflicted {damage} {skill_name} damage ' +
                        f'(attack equip lvl {equipment_level_offense} vs ' +
                        f'defense equip lvl {equipment_level_defense})',
                        tags={"player_id": player.ent_id})

  player.apply_damage(damage, skill.__class__.__name__.lower())
  target.receive_damage(player, damage)

  return damage


def danger(config, pos):
  border = config.MAP_BORDER
  center = config.MAP_CENTER
  r, c   = pos

  #Distance from border
  r_dist  = min(r - border, center + border - r - 1)
  c_dist  = min(c - border, center + border - c - 1)
  dist   = min(r_dist, c_dist)
  norm   = 2 * dist / center

  return norm

def spawn(config, dnger, np_random):
  border = config.MAP_BORDER
  center = config.MAP_CENTER
  mid    = center // 2

  dist       = dnger * center / 2
  max_offset = mid - dist
  offset     = mid + border + np_random.integers(-max_offset, max_offset)

  rng = np_random.random()
  if rng < 0.25:
    r = border + dist
    c = offset
  elif rng < 0.5:
    r = border + center - dist - 1
    c = offset
  elif rng < 0.75:
    c = border + dist
    r = offset
  else:
    c = border + center - dist - 1
    r = offset

  if __debug__:
    assert dnger == danger(config, (r,c)), 'Agent spawned at incorrect radius'

  r = int(r)
  c = int(c)

  return r, c



# pylint: disable=cyclic-import
from nmmo.core import action
from nmmo.systems.ai import utils

DIRECTIONS = [ # row delta, col delta, action
      (-1, 0, action.North),
      (1, 0, action.South),
      (0, -1, action.West),
      (0, 1, action.East)] * 2

def habitable(realm_map, ent, np_random):
  r, c = ent.pos
  is_habitable = realm_map.habitable_tiles
  start = np_random.get_direction()
  for i in range(4):
    dr, dc, act = DIRECTIONS[start + i]
    if is_habitable[r + dr, c + dc]:
      return act

  return action.North

def towards(direction, np_random):
  if direction == (-1, 0):
    return action.North
  if direction == (1, 0):
    return action.South
  if direction == (0, -1):
    return action.West
  if direction == (0, 1):
    return action.East

  return np_random.choice(action.Direction.edges)

def bullrush(ent, targ, np_random):
  direction = utils.directionTowards(ent, targ)
  return towards(direction, np_random)

def pathfind(realm_map, ent, targ, np_random):
  direction = utils.aStar(realm_map, ent.pos, targ.pos)
  return towards(direction, np_random)

def antipathfind(realm_map, ent, targ, np_random):
  er, ec = ent.pos
  tr, tc = targ.pos
  goal   = (2*er - tr , 2*ec-tc)
  direction = utils.aStar(realm_map, ent.pos, goal)
  return towards(direction, np_random)



import numpy as np
from scipy import signal

from nmmo.lib.colors import Neon

# NOTE: added to fix json.dumps() cannot serialize numpy objects
# pylint: disable=inconsistent-return-statements
def np_encoder(obj):
  if isinstance(obj, np.generic):
    return obj.item()

def normalize(ary: np.ndarray, norm_std=2):
  R, C         = ary.shape
  preprocessed = np.zeros_like(ary)
  nonzero      = ary[ary!= 0]
  mean         = np.mean(nonzero)
  std          = np.std(nonzero)
  if std == 0:
    std = 1
  for r in range(R):
    for c in range(C):
      val = ary[r, c]
      if val != 0:
        val = (val - mean) / (norm_std * std)
        val = np.clip(val+1, 0, 2)/2
        preprocessed[r, c] = val
  return preprocessed

def clip(ary: np.ndarray):
  R, C         = ary.shape
  preprocessed = np.zeros_like(ary)
  nonzero      = ary[ary!= 0]
  mmin         = np.min(nonzero)
  mmag         = np.max(nonzero) - mmin
  for r in range(R):
    for c in range(C):
      val = ary[r, c]
      val = (val - mmin) / mmag
      preprocessed[r, c] = val
  return preprocessed

def make_two_tone(ary, norm_std=2, preprocess='norm', invert=False, periods=1):
  if preprocess == 'norm':
    ary   = normalize(ary, norm_std)
  elif preprocess == 'clip':
    ary   = clip(ary)

  # if preprocess not in ['norm', 'clip'], assume no preprocessing
  R, C      = ary.shape

  colorized = np.zeros((R, C, 3))
  if periods != 1:
    ary = np.abs(signal.sawtooth(periods*3.14159*ary))
  if invert:
    colorized[:, :, 0] = ary
    colorized[:, :, 1] = 1-ary
  else:
    colorized[:, :, 0] = 1-ary
    colorized[:, :, 1] = ary

  colorized *= (ary != 0)[:, :, None]

  return colorized

# TODO: this is a hack to make the client work
#   by adding color, population, self to the packet
#   integrating with team helper could make this neat
def patch_packet(packet, realm):
  for ent_id in packet['player']:
    packet['player'][ent_id]['base']['color'] = Neon.GREEN.packet()
    # EntityAttr: population was changed to npc_type
    packet['player'][ent_id]['base']['population'] = 0
    # old code: nmmo.Serialized.Entity.Self, no longer being used
    packet['player'][ent_id]['base']['self'] = 1

  npc_colors = {
    1: Neon.YELLOW.packet(), # passive npcs
    2: Neon.MAGENTA.packet(), # neutral npcs
    3: Neon.BLOOD.packet() } # aggressive npcs
  for ent_id in packet['npc']:
    npc = realm.npcs.corporeal[ent_id]
    packet['npc'][ent_id]['base']['color'] = npc_colors[int(npc.npc_type.val)]
    packet['npc'][ent_id]['base']['population'] = -int(npc.npc_type.val) # note negative
    packet['npc'][ent_id]['base']['self'] = 1

  return packet
